# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import logging

from pydoop.pipes import Reducer
from pydoop.utils import jc_configure, jc_configure_bool

import bl.lib.seq.aligner.io.protobuf_mapping as protobuf_mapping
from bl.mr.lib.hadoop_event_monitor import HadoopEventMonitor
from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink
from bl.mr.lib.emit_sam_link import EmitSamLink
import bl.lib.tools.deprecation_utils as deprecation_utils
import seqal_app

from bl.lib.seq.aligner.io.sam_formatter import SamFormatter

class reducer(Reducer):
	COUNTER_CLASS = "SEQAL" # TODO:  refactor so that mapper and reducer have a common place for things like this constant

	DeprecationMap = {
	  'seal.seqal.log.level': 'bl.seqal.log.level',
	  'seal.seqal.discard_duplicates': 'bl.seqal.discard_duplicates'
	}

	def __init__(self, ctx):
		super(reducer, self).__init__(ctx)

		jc = ctx.getJobConf()
		logger = logging.getLogger("seqal")
		jobconf = deprecation_utils.convert_job_conf(jc, self.DeprecationMap, logger)

		jc_configure(self, jobconf, 'seal.seqal.log.level', 'log_level', 'INFO')
		jc_configure_bool(self, jobconf, 'seal.seqal.discard_duplicates', 'discard_duplicates', False)

		logging.basicConfig(level=self.log_level)

		self.event_monitor = HadoopEventMonitor(self.COUNTER_CLASS, logging.getLogger("reducer"), ctx)
		self.__output_sink = EmitSamLink(ctx, self.event_monitor)

	def __process_unmapped_pairs(self, ctx):
		while ctx.nextValue():
			value = ctx.getInputValue()
			pair = protobuf_mapping.unserialize_pair(value)
			self.__output_sink.process(pair)

	def reduce(self, ctx):
		# create the "workspace"
		self.__pairs = []
		self.__unpaired = []

		# gather input
		key_values = ctx.getInputKey().split(':')

		if key_values[0] == seqal_app.UNMAPPED_STRING:
			# pair of unmapped sequences
			self.__process_unmapped_pairs(ctx)
		else:
			if len(key_values) != 3:
				raise RuntimeError("Unexpected key length %d.  Expected key format is ref_id:pos:orient" % len(key))
			# convert key values and make it a tuple
			key = (int(key_values[0]), int(key_values[1]), key_values[2] == 'R') # last value is True if reverse strand

			have_pairs = False # keep track of whether we have at least one real pair.
			# load mappings
			while ctx.nextValue():
				value = ctx.getInputValue()
				if value == seqal_app.PAIR_STRING:
					have_pairs = True
				else:
					pair = protobuf_mapping.unserialize_pair(value)
					if pair[0] is None or pair[0].is_unmapped():
						# Sanity check. pair[0] should never be None or unmapped here.
						raise ValueError("Error!  Got None or unmapped in first read for key %s.  pair: %s" % (key, pair))

					if pair[1] and pair[1].is_unmapped():
						self.__output_sink.process( (pair[1], None) )
						self.__unpaired.append( (pair[0], None) )
					elif pair[1] is None:
						self.__unpaired.append(pair)
					else:
						# Two mapped reads.
						# pair[0] should never be unmapped.  That case should be handled by
						# __process_unmapped_pairs.
						self.__pairs.append(pair)
						have_pairs = True

			self.__process_pairs()
			self.__process_fragments(have_pairs)

		# clean-up the workspace
		self.__pairs = None
		self.__unpaired = None

	def __process_pairs(self):
		# All pairs whose 5'-most coordinate matches the key,
		# and are not duplicate pairs, will be emitted
		keep_pairs = dict()
		for p in self.__pairs:
			p_key = get_pair_key(p) # makes the key on which we base the comparison between pairs

			# If we already have a pair with this key, then keep the one with the highest score.
			# If we haven't already seen the key, put the pair in the hash.
			if keep_pairs.has_key(p_key):
				if get_map_pair_score(keep_pairs[p_key]) < get_map_pair_score(p):
					dup_pair = keep_pairs[p_key]
					keep_pairs[p_key] = p
				else:
					dup_pair = p
				self.event_monitor.count("duplicate pairs")
				if not self.discard_duplicates: # emit the duplicates if we need to
					for r in dup_pair:
						r.set_duplicate(True)
					self.__output_sink.process(dup_pair)
			else:
				keep_pairs[p_key] = p
		# finally, emit the pairs that we've kept
		self.event_monitor.count("rmdup unique pairs", len(keep_pairs))
		for pair in keep_pairs.itervalues():
			self.__output_sink.process(pair)

	def __process_fragments(self, with_pairs):
		# All fragments that are not the duplicate of another
		# fragment, be it in a pair or alone, will be emitted.
		#
		# All fragments we analyze here will have been emitted for the same coordinate
		# (the one referenced by the key).  Therefore, they automatically have a
		# duplicate in any pairs we have received. As a consequence, we only look at
		# them if we haven't seen any pairs.
		#
		# with_pairs => implies we have proper pairs for the key position,
		# so all lone fragments are to be discarded as duplicates.
		#
		# not with_pairs => we have no proper pairs for the key position.
		# Duplicates will be selected by quality."""
		if with_pairs:
			# all fragments are duplicates
			self.event_monitor.count("duplicate fragments", len(self.__unpaired))
			if not self.discard_duplicates:
				for dup in self.__unpaired: # for each unpaired fragment
					dup[0].set_duplicate(True)
					self.__output_sink.process(dup)
		else:
			fragments = dict()
			for m,none in self.__unpaired: # for each unpaired fragment
				k = get_mapping_key(m)
				if fragments.has_key(k):
					if get_map_score(fragments[k]) < get_map_score(m):
						dup = fragments[k]
						fragments[k] = m
					else:
						dup = m
					self.event_monitor.count("duplicate fragments")
					if not self.discard_duplicates:
						dup.set_duplicate(True)
						self.__output_sink.process((dup,None))
				else:
					fragments[k] = m
			# now emit the remaining fragments
			self.event_monitor.count("rmdup unique fragments", len(fragments))
			for m in fragments.itervalues():
				self.__output_sink.process( (m,None) )

def get_mapping_key(mapping):
	if mapping.is_on_reverse():
		pos = mapping.get_untrimmed_right_pos()
		onreverse = True
	else:
		pos = mapping.get_untrimmed_left_pos()
		onreverse = False
	return (mapping.ref_id, pos, onreverse)

def get_pair_key(pair):
	return (
	  pair[0].ref_id,
	  pair[0].get_untrimmed_right_pos() if pair[0].is_on_reverse() else pair[0].get_untrimmed_left_pos(),
	  pair[0].is_on_reverse(),
	  pair[1].ref_id,
	  pair[1].get_untrimmed_right_pos() if pair[1].is_on_reverse() else pair[1].get_untrimmed_left_pos())

def get_map_score(mapping):
	"""
	Sum of all base quality scores >= 15.
	"""
	bq = mapping.get_base_qualities()
	return sum([value for value in bq if value >= 15]) if bq else 0

def get_map_pair_score(pair):
	return get_map_score(pair[0]) + get_map_score(pair[1])



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

"""
How to use this test:

You need to provide as command line arguments:
  1. The fasta reference sequence file name
  2, 3. The two reads and mates sequence files

use LOG_LEVEL = logging.DEBUG; N_ITER = 1 to see details.
use LOG_LEVEL = logging.INFO; N_ITER = 10 to check for leaks.
"""

import sys, gc, logging
#print gc.isenabled()
gc.set_debug(gc.DEBUG_UNCOLLECTABLE)

import itertools as it
import bl.lib.seq.aligner.bwa as bwa

import Bio.SeqIO
from bl.lib.util.meminfo import meminfo


MB = float(2**20)


def print_meminfo(status, log_function=logging.debug):
  minfo = meminfo()
  msg = "%s: size = %.1fm; resident = %.1fm" % (
    status, minfo["size"]/MB, minfo["resident"]/MB)
  log_function(msg)


def run_bwa_py_sampe_alloc_only(refseq_fname, read_fname, mate_fname):
  read_flow = Bio.SeqIO.parse(open(read_fname), 'fastq-illumina')
  mate_flow = Bio.SeqIO.parse(open(mate_fname), 'fastq-illumina')
  pairs = [x for x in it.izip(read_flow, mate_flow)]
  print_meminfo("AFTER READING PAIRS")
  bwsa = bwa.build_bws_array(pairs)
  print_meminfo("AFTER BUILDING BWSA")
  bwts = bwa.restore_index(refseq_fname)
  print_meminfo("AFTER RESTORING INDEX")
  bnsp, pacseq = bwa.restore_reference(refseq_fname)
  print_meminfo("AFTER RESTORING REFERENCE")
  gopt, popt = bwa.gap_init_opt(), bwa.pe_init_opt()
  ii, last_ii = bwa.isize_info_t(), bwa.isize_info_t()
  last_ii.avg = -1.0
  l = len(pairs)
  print_meminfo("AFTER INIT OPT & II")
  # deallocate seq & ref data
  for i in 0, 1:
    bwa.free_seq(l, bwsa[i])
    bwa.bwt_destroy(bwts[i])
  bwa.bns_destroy(bnsp)
  print_meminfo("AFTER DEALLOC")
  del pacseq
  n_unreachable = gc.collect()
  logging.debug("n_unreachable = %d" % n_unreachable)
  print_meminfo("AFTER DEL PACSEQ")
  del pairs
  n_unreachable = gc.collect()
  logging.debug("n_unreachable = %d" % n_unreachable)
  print_meminfo("AFTER DEL PAIRS")


def run_bwa_py_sampe(refseq_fname, read_fname, mate_fname):
  read_flow = Bio.SeqIO.parse(open(read_fname), 'fastq-illumina')
  mate_flow = Bio.SeqIO.parse(open(mate_fname), 'fastq-illumina')
  pairs = [x for x in it.izip(read_flow, mate_flow)]
  print_meminfo("AFTER READING PAIRS")
  bwsa = bwa.build_bws_array(pairs)
  print_meminfo("AFTER BUILDING BWSA")
  bwts = bwa.restore_index(refseq_fname)
  print_meminfo("AFTER RESTORING INDEX")
  bnsp, pacseq = bwa.restore_reference(refseq_fname)
  print_meminfo("AFTER RESTORING REFERENCE")
  gopt, popt = bwa.gap_init_opt(), bwa.pe_init_opt()
  ii, last_ii = bwa.isize_info_t(), bwa.isize_info_t()
  last_ii.avg = -1.0
  l = len(pairs)
  print_meminfo("AFTER INIT OPT & II")
  bwa.cal_sa_reg_gap(0, bwts, l, bwsa[0], gopt)
  bwa.cal_sa_reg_gap(0, bwts, l, bwsa[1], gopt)
  print_meminfo("AFTER CAL_SA_REG_GAP")
  cnt_chg = bwa.cal_pac_pos_pe(bwts, l, bwsa, ii, popt, gopt, last_ii)
  print_meminfo("AFTER CAL_PAC_POS_PE")
  bwa.paired_sw(bnsp, pacseq, l, bwsa, popt, ii)
  print_meminfo("AFTER PAIRED_SW")
  bwa.refine_gapped(bnsp, l, bwsa[0], pacseq)
  bwa.refine_gapped(bnsp, l, bwsa[1], pacseq)
  print_meminfo("AFTER REFINE_GAPPED")
  for k in xrange(l):
    v1 = bwa.analyze_hit(gopt[0], bnsp, bwsa[0][k], bwsa[1][k])
    v2 = bwa.analyze_hit(gopt[0], bnsp, bwsa[1][k], bwsa[0][k])
  print_meminfo("AFTER ANALYZE_HIT")
  # deallocate seq & ref data
  for i in 0, 1:
    bwa.free_seq(l, bwsa[i])
    bwa.bwt_destroy(bwts[i])
  bwa.bns_destroy(bnsp)
  print_meminfo("AFTER DEALLOC")
  del pacseq
  n_unreachable = gc.collect()
  logging.debug("n_unreachable = %d" % n_unreachable)
  print_meminfo("AFTER DEL PACSEQ")
  del pairs
  n_unreachable = gc.collect()
  logging.debug("n_unreachable = %d" % n_unreachable)
  print_meminfo("AFTER DEL PAIRS")


def main(argv):
  LOG_LEVEL = logging.DEBUG; N_ITER = 1
  #LOG_LEVEL = logging.INFO; N_ITER = 10
  logging.basicConfig(level=LOG_LEVEL)
  #fun = run_bwa_py_sampe_alloc_only
  fun = run_bwa_py_sampe

  print_meminfo("START", logging.info)

  try:
    refseq_fn = argv[1]
    read_fn = argv[2]
    mate_fn = argv[3]
  except IndexError:
    sys.exit("Usage: %s REFSEQ_FN READ_FN MATE_FN" % sys.argv[0] + __doc__)
  #u.build_index(refseq_fn)
  for i in xrange(N_ITER):
    fun(refseq_fn, read_fn, mate_fn)
    print_meminfo("END ITERATION %d" % i, logging.info)


if __name__ == "__main__":
  main(sys.argv)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Seal is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Seal.  If not, see <http://www.gnu.org/licenses/>.

# a couple of utility functions to deal with deprecated properties

from bl.mr.lib.jc_wrapper import jc_wrapper

# Warn the user that he's using a deprecated property.
# If new_property is provided, the message will suggest to the user to substitute uses
# of the deprecatedProperty with the new one.
def deprecation_warning(log, deprecated_property, new_property):
	log.warning("Your configuration is using the deprecated property %s", deprecated_property)
	if new_property is None:
		log.warning("You should update your configuration to avoid using it.  See the documentation for details.")
	else:
		log.warning("You should update your configuration to replace it with %s. See the documentation for details.", new_property)


# Check whether a deprecated property is used, and if so emit a warning.
# @param job_conf A Pydoop JobConf object.
# @param deprecatedProperty The name of the deprecated property.
# @param new_property The name of the property that replaces the deprecated one, if any,
#        to write a suggestion to the user.
def check_deprecated_prop(job_conf, log, deprecated_property, new_property):
	if job_conf.hasKey(deprecated_property):
		deprecation_warning(log, deprecated_property, new_property)


def convert_job_conf(jobconf, deprecation_map, logger):
	wrapper = jc_wrapper(jobconf)

	for new, old in deprecation_map.iteritems():
		check_deprecated_prop(wrapper, logger, old, new)
		if wrapper.hasKey(old):
			wrapper[new] = wrapper[old]
			logger.warning("Using value %s for property %s (value taken from its deprecated equivalent property %s).", wrapper[new], new, old)
	return wrapper



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import sys, itertools as it
import Bio.SeqIO
import logging
logging.basicConfig(level=logging.DEBUG)

import bl.lib.seq.aligner.bwa as bwa
from bl.lib.seq.aligner.bwa.bwa_iterator import BWAIterator, MRVisitor


def get_counters():
  counter_names = [
    "restore_index",
    "restore_reference",
    "cal_sa_reg_gap",
    "cal_pac_pos_pe",
    "paired_sw",
    "paired_sw_batch",
    "refine_gapped"]
  counters = {}
  for cn in counter_names:
    counters[cn] = CounterStub()
  return counters


class CounterStub(object):

  def __init__(self, value=0):
    self.value = value

  def increment(self, inc):
    self.value += inc


class ContextStub(object):

  def incrementCounter(self, counter, inc):
    counter.increment(inc)

  def setStatus(self, status):
    sys.stderr.write("Status set to '%s'\n" % status)


def main(argv):

  try:
    refseq_fname = argv[1]
    read_fname = argv[2]
    mate_fname = argv[3]
  except IndexError:
    sys.exit("Usage: %s REFSEQ_FN READ_FN MATE_FN" % sys.argv[0])

  seq_list_len = 5000
  max_isize = pairing_batch_size = 1000
  gopt, popt = bwa.gap_init_opt(), bwa.pe_init_opt()

  read_flow = Bio.SeqIO.parse(open(read_fname), 'fastq-illumina')
  mate_flow = Bio.SeqIO.parse(open(mate_fname), 'fastq-illumina')
  pairs_flow = it.izip(read_flow, mate_flow)
  res = []
  while 1:
    pairs = list(it.islice(pairs_flow, 0, seq_list_len))
    if len(pairs) == 0:
      break
    bwts = bwa.restore_index(refseq_fname)
    bnsp, pacseq = bwa.restore_reference(refseq_fname)

    l = len(pairs)
    bwsa = bwa.build_bws_array(pairs)

    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    counters = get_counters()
    ctx = ContextStub()
    visitor = MRVisitor(logger, ctx, counters)

    bwa_iterator = BWAIterator(refseq_fname, gopt, popt, max_isize,
                               pairing_batch_size, visitor)
    for read, mate in bwa_iterator.analyze(bwsa, l):
      print read.get_name(), mate.get_name()

    for j in 0, 1:
      bwa.free_seq(l, bwsa[j])
    bwa.bns_destroy(bwa_iterator.bnsp)

  for cn, c in counters.iteritems():
    sys.stderr.write("%s = %d\n" % (cn, c.value))


if __name__ == "__main__":
  main(sys.argv)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

class SamFormatter(object):
  """
  Creates SAM records from BwaAlignment objects.
  """

  SAM_FIELDS = [
    'qname',
    'flag',
    'rname',
    'pos',
    'mapq',
    'cigar',
    'mrnm',
    'mpos',
    'isize',
    'seq',
    'qual'
    ]

  def __init__(self, strip_pe_tag=True):
    self.strip_pe_tag = strip_pe_tag

  def format(self, mapping):
    sam_record = {}
    sam_record['qname'] = mapping.get_name()
    if self.strip_pe_tag:
      sam_record['qname'] = sam_record['qname'].rsplit("/",1)[0]
    sam_record['flag'] = str(mapping.flag)
    sam_record['rname'] = mapping.tid or '*'
    sam_record['pos'] = str(mapping.pos)
    sam_record['mapq'] = str(mapping.qual)
    sam_record['cigar'] = mapping.get_cigar_str()
    sam_record['mrnm'] = mapping.mtid or '*'
    sam_record['mpos'] = str(mapping.mpos)
    sam_record['isize'] = str(mapping.isize)
    sam_record['seq'] = mapping.get_seq_5()
    sam_record['qual'] = mapping.get_ascii_base_qual()
    sam_aux = []
    for tag in mapping.each_tag():
      sam_aux.append('%s:%s:%s' % (tag[0], tag[1], tag[2]))
    sam_record['aux'] = sam_aux
    sam_record_str = '\t'.join([sam_record[k] for k in self.SAM_FIELDS])
    if sam_record['aux']:
      sam_record_str += '\t' + '\t'.join(sam_record['aux'])
    return sam_record_str

  def parse(self, sam_record_str):
    r = sam_record_str.split('\t')
    return dict(name=r[0],
                flag=int(r[1]),
                tid=r[2],
                pos=int(r[3]),
                qual=int(r[4]),
                cigar=self.__convert_cigar(r[5]),
                mtid=r[6],
                mpos=int(r[7]),
                isize=int(r[8]),
                seq=r[9],
                qualseq=r[10],
                aux=self.__convert_aux(r[11:]))

  def __convert_cigar(self, cig):
    res, d = [], ''
    for c in cig:
      if c.isalpha():
        res.append((int(d), c))
        d = ''
      else:
        d += c
    return res

  def __convert_aux(self, aux):
    def convert(x):
      return (x[0], x[1],
              int(x[2]) if x[2].isdigit() else x[2])
    return [convert(x.split(':')) for x in aux]



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

# FIXME: this does not belong here. We should move it to pydoop or,
# even better, add a testing utilities framework to pydoop (see ticket
# #277) and use that.

#from pydoop._pipes import MapContext, ReduceContext
import StringIO

class SavingLogger(object):
	DEBUG = 50
	INFO = 40
	WARNING = 30
	ERROR = 20
	CRITICAL = 10

	def __init__(self):
		self.output = StringIO.StringIO()
		self.log_level = type(self).DEBUG

	def contents(self):
		return self.output.getvalue()

	def debug(self, *args):
		if self.log_level >= type(self).DEBUG:
			self.output.write(' '.join(map(str, args)))
			self.output.write("\n")

	def info(self, *args):
		if self.log_level >= type(self).INFO:
			self.output.write(' '.join(map(str, args)))
			self.output.write("\n")

	def warning(self, *args):
		if self.log_level >= type(self).WARNING:
			self.output.write(' '.join(map(str, args)))
			self.output.write("\n")

	def error(self, *args):
		if self.log_level >= type(self).ERROR:
			self.output.write(' '.join(map(str, args)))
			self.output.write("\n")

	def critical(self, *args):
		if self.log_level >= type(self).CRITICAL:
			self.output.write(' '.join(map(str, args)))
			self.output.write("\n")

class map_context(object):

  def __init__(self, jc, input_split):
    self.job_conf = jc
    self.input_split = input_split
    self.counters = {}
    self.emitted = {}
    self.status_messages = []

  def getJobConf(self):
    return self.job_conf

  def getInputSplit(self):
    return self.input_split

  def getInputKey(self):
    return self.input_key

  def getInputValue(self):
    return self.input_value

  def getCounter(self, group, name):
    k = '%s:%s' % (group, name)
    self.counters[k] = 0
    return k

  def incrementCounter(self, counter, amount):
    self.counters[counter] += amount

  def emit(self, k, v):
    if not isinstance(k, str):
      raise TypeError("key must be a string (it's a %s)" % type(k))
    if not isinstance(v, str):
      raise TypeError("value must be a string (it's a %s)" % type(v))

    if not self.emitted.has_key(k):
      self.emitted[k] = []
    self.emitted[k].append(v)

  def progress(self):
    pass

  def setStatus(self, status):
    self.status_messages.append(status)


class reduce_context(object):
  # @param jc:  JobConf object
  # @param values:  a list of dictionaries. The first dictionary
  #          needs to have a 'key' and a 'value'.  Each subsequent
  #          only needs a 'value'.
  def __init__(self, jc, values):
    self.job_conf = jc
    self.counters = {}
    self.emitted = {}
    self.counter = -1
    self.values = values

  def nextValue(self):
    self.counter += 1
    return self.counter < len(self.values)

  def getJobConf(self):
    return self.job_conf

  def getInputKey(self):
    return self.values[0]['key']

  def getInputValue(self):
    return self.values[self.counter]['value']

  def emit(self, k, v):
    if not isinstance(k, str):
      raise TypeError("key must be a string (it's a %s)" % type(k))
    if not isinstance(v, str):
      raise TypeError("value must be a string (it's a %s)" % type(v))

    if not self.emitted.has_key(k):
      self.emitted[k] = []
    self.emitted[k].append(v)

  def progress(self):
    pass

  def setStatus(self, status):
    pass

  def getCounter(self, group, name):
    k = '%s:%s' % (group, name)
    self.counters[k] = 0
    return k

  def incrementCounter(self, counter, amount):
    self.counters[counter] += amount

  def add_value(self, key, value):
    """
    Convenience method to insert a (key, value) pair to the data that
    will be returned by this reduce_context.
    It is recommended that this method be used in place of directly
    modifying the object's 'value' attribute so that in the future
    we may easily change the internals.
    """
    if self.values:
      if self.values[0]['key'] == key:
        self.values.append({'value': value })
      else:
        raise ValueError("key %s doesn't match the key that's already been inserted (%s).  " +
                         "Sorry, but for now we only support a single key value" % (key, self.values[0]['key']))
    else: # empty values
      self.values.append({'key':key, 'value':value})



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

from constants import *
from bl.lib.seq.aligner.sam_flags import *
from bl.lib.seq.aligner.mapping import Mapping
import bwa_core as bwa

import sys
import array
from ctypes import *

class BwaMapping(Mapping):
  """
  Specialization of the seq.aligner.Mapping class to work with bwa hits. Note that the constructor expects
  its arguments to be actual objects, not pointers to them. Example:

    >>> gopt = bwa.gap_init_opt()
    >>> bnsp, pacseq = bwa.restore_reference(root)
    >>> hw = BwaMapping(gopt[0], bnsp[0], read, mate)

    >>> res = bwa.analyze_hit(gopt[0], bnsp, read, mate)
    >>> print res['name']
    foo-00000/1
    >>> print hw.name
    foo-00000/1
  """
  def __init__(self, opts, bns, read, mate=None):
    """
    @param opts:  bwa parameter structure gap_opt_t
    @param bns:  reference structure bntseq_t
    @param read: bwa_seq_t
    @param mate: bwa_seq_t

    Added attributes:
    self.bns: reference structure
    self.opts: the BWA options structure provided on construction
    """
    super(BwaMapping, self).__init__()
    #self.print_debug_info(self.__read)
    self.opts = opts
    self.bns = bns
    self.__read = read
    self.__mate = mate
    self.flag = self.__read.extra_flag

    if (read.type != BWA_TYPE_NO_MATCH or (mate and mate.type != BWA_TYPE_NO_MATCH)):
      self.__build_hit_for_match()
    else:
     # both read and mate are NO_MATCH
      self.flag |= SAM_FSU
      if self.__mate:
        self.flag |= SAM_FMU

  def __build_hit_for_match(self):
    """At least one of read or mate are a match"""
    if self.__read.type == BWA_TYPE_NO_MATCH:
      # the read isn't matched, which implies we have a mate
      # and it is matched.  We place the read at the mate's position
      self.pos = self.__mate.pos
      strand = self.__mate.strand
      self.flag |= SAM_FSU # query unmapped
      ref_len = 1 # ref_len is the length of the reference in the alignment
    else:
      self.pos = self.__read.pos
      strand = self.__read.strand
      ref_len = self.__read.get_pos_end() - self.pos
      self.qual = self.__read.mapQ

    if strand:
      self.flag |= SAM_FSR # query on the reverse strand

    # Where in the reference is pos?
    self.ref_id, self.__nn = bwa.get_seq_id(self.bns, self.pos, ref_len)
    self.tid = self.bns.anns[self.ref_id].name
    # make the position relative to the start of its reference section
    self.pos = self.pos - self.bns.anns[self.ref_id].offset

    # Check whether the read alignment bridges two adjacent reference sequences
    if (self.__read.type != BWA_TYPE_NO_MATCH and
        (self.pos + ref_len > self.bns.anns[self.ref_id].len)):
      self.flag |= SAM_FSU # it does, so we set it as unmapped

    # presume we can't calculate any of these fields,
    # so we set them to the "default" values"
    self.mtid = None
    self.mpos = 0
    self.isize = 0

    if self.__mate:
      if self.__mate.type != BWA_TYPE_NO_MATCH:
        if self.__mate.strand:
          self.flag |= SAM_FMR
        self.m_ref_id, m_nn = bwa.get_seq_id(self.bns, self.__mate.pos, self.__mate.len)
        if self.m_ref_id == self.ref_id:
          self.mtid = '='
          if self.__read.type != BWA_TYPE_NO_MATCH:
            self.isize = self.__mate.get_pos_5() - self.__read.get_pos_5()
            # else the isize remains at 0
        else:
          self.mtid = self.bns.anns[self.m_ref_id].name
        self.mpos = self.__mate.pos - self.bns.anns[self.m_ref_id].offset
      else: # unmapped mate
        self.flag |= SAM_FMU
        self.mtid = "="
        self.mpos = self.__read.pos - self.bns.anns[self.ref_id].offset

    self.pos = self.pos + 1 # base 1
    self.mpos = self.mpos + 1 # base 1

  def get_name(self):
    if not hasattr(self, "__name"):
      self.__name = self.__read.get_name()
    return self.__name

  def get_base_qualities(self):
    if not hasattr(self, '__base_qualities'):
      if self.__read.qual:
        # BWA keeps base quality is in ASCII-33
        it = xrange(self.__read.full_len-1,-1,-1) if self.is_on_reverse() else xrange(self.__read.full_len)
        try:
          self.__base_qualities = array.array("B", [ self.__read.qual[i] - 33 for i in it ])
        except OverflowError:
          raise ValueError("Base quality value out of range.  Have you set the base quality format correctly? (property name: seal.seqal.fastq-subformat)")
      else:
        self.__base_qualities = None
    return self.__base_qualities

  def get_cigar(self):
    if self.is_mapped():
      return self.__read.get_cigar()
    else:
      return []

  ####################################
  # Tag methods
  ####################################
  def each_tag(self):
    if not hasattr(self, '__tags'):
      self.__compute_tags()
    for tpl in self.__tags:
      yield tpl

  def __find_tag(self, name):
    if not hasattr(self, '__tags'):
      self.__compute_tags()
    for tpl in self.__tags:
      if name == tpl[0]:
        return tpl
    return None

  def tag_value(self, tag_name):
    """
    Get the value for a tag, if present.
    Returns:
      - integer (if tag type is 'i')
      - float (if tag type is 'f')
      - string
      - None, if tag isn't present
    """
    tag = self.__find_tag(tag_name) # calls __compute_tags if necessary
    if tag:
      if tag[1] == 'i':
        value = int(tag[2])
      elif tag[1] == 'f':
        value = float(tag[2])
      else:
        value = tag[2]
      return value
    else:
      return None

  def __compute_tags(self):
    """
    Tags starting with 'X' are specific to BWA.

    +-----+---------------------------------------------------------------+
    | Tag |                    Meaning                                    |
    +-----+---------------------------------------------------------------+
    | NM  | Edit distance (number of nucleotide differences)              |
    | MD  | Mismatching positions/bases                                   |
    | AS  | Alignment score                                               |
    | CM  | Number of color differences                                   |
    | SM  | Mapping quality if read is mapped as single read              |
    | AM  | Smaller single-end mapping quality of the two reads in a pair |
    +-----+---------------------------------------------------------------+
    | X0  | Number of best hits                                           |
    | X1  | Number of suboptimal hits found by BWA                        |
    | XN  | Number of ambiguous bases in the referenece                   |
    | XM  | Number of mismatches in the alignment                         |
    | XO  | Number of gap opens                                           |
    | XG  | Number of gap extentions                                      |
    | XT  | Type: Unique/Repeat/N/Mate-sw                                 |
    | XA  | Alternative hits; format: (chr,pos,CIGAR,NM;)*                |
    | XS  | Suboptimal alignment score                                    |
    | XF  | Support from forward/reverse alignment                        |
    | XE  | Number of supporting seeds                                    |
    | XC  | Length of clipped reads                                       |
    +-----+---------------------------------------------------------------+

    Note that XO and XG are generated by BWT search while the CIGAR
    string by Smith-Waterman alignment. These two tags may be
    inconsistent with the CIGAR string. This is not a bug.
    """
    self.__tags = []
    if (self.__read.clip_len < self.__read.full_len):
      self.__tags.append(("XC", "i", self.__read.clip_len))

    if self.__read.type != BWA_TYPE_NO_MATCH:
      # min_seQ: the smaller of the two mapping qualities
      if self.__mate and self.__mate.type != BWA_TYPE_NO_MATCH:
        min_seQ = min(self.__read.seQ, self.__mate.seQ)
      else:
        min_seQ = 0 # else the mate isn't mapped and therefore 0

      mode, max_top2 = self.opts.mode, self.opts.max_top2
      XT = "NURM"[self.__read.type] if self.__nn <=10 else 'N'
      self.__tags.append(('XT', 'A', XT))
      self.__tags.append(('NM' if mode & BWA_MODE_COMPREAD else 'CM',
                   'i', self.__read.nm))
      if self.__nn:
        self.__tags.append(('XN', 'i', self.__nn))
      if self.__mate:
        self.__tags.append(('SM', 'i', self.__read.seQ))
        self.__tags.append(('AM', 'i', min_seQ))
      if self.__read.type != BWA_TYPE_MATESW:
        #NOTE this is X0 (digit 0  NOT the letter O)
        self.__tags.append(('X0', 'i', self.__read.c1))
        if self.__read.c1 < max_top2:
          self.__tags.append(('X1', 'i', self.__read.c2))
      self.__tags.append(('XM', 'i', self.__read.n_mm))
      #NOTE this is XO (letter O NOT the digit 0)
      self.__tags.append(('XO', 'i', self.__read.n_gapo))
      self.__tags.append(('XG', 'i', self.__read.n_gapo + self.__read.n_gape))
      if self.__read.md:
        self.__tags.append(('MD', 'Z', self.__read.md))
      # FIXME: for consistency, the third field of the XA tag should be a
      # list of tuples to be converted to a string upon formatting.
      if self.__read.n_multi:
        multi_hit = []
        for i in xrange(self.__read.n_multi):
          multi_hit_chunk = self.__analyze_multi(i)
          multi_hit.append(",".join(multi_hit_chunk))
        self.__tags.append(('XA', 'Z', ";".join(multi_hit)+";"))

  def __analyze_multi(self, i):
    multi_hit_chunk = []
    q = self.__read.multi[i]
    ref_len = q.get_pos_end(len(self.__read)) - q.pos
    ref_id, nn = bwa.get_seq_id(self.bns, q.pos, ref_len)
    sign = '-' if q.strand else '+'
    multi_hit_chunk.append(self.bns.anns[ref_id].name)
    multi_hit_chunk.append("%s%d" % (sign, (q.pos - self.bns.anns[ref_id].offset + 1)))
    # FIXME: duplicate from samt_adapter, we should centralize formatting
    cigar_str = "".join(['%d%s' % t for t in q.get_cigar(len(self.__read))]) or "*"
    multi_hit_chunk.append(cigar_str)
    multi_hit_chunk.append("%d" % (q.n_gapo + q.n_gape + q.n_mm))
    return multi_hit_chunk

  def get_seq_5(self):
    if not hasattr(self, "__seq5"):
      if self.is_on_reverse():
        # The sequence self.__read.get_rseq()[::-1] is shifted by one position.  The way to get
        # the correct reverse complement sequence is to build it from seq, which is exactly
        # what BWA normally does.
        self.__seq5 = ''.join([ "TGCAN"[self.__read.seq[k]] for k in xrange(self.__read.full_len-1, -1, -1)])
      else:
        self.__seq5 = self.__read.get_seq()
    return self.__seq5

  def get_seq_len(self):
    return self.__read.full_len

  def __str__(self):
    return "%s\tflags: %s" % (self.get_name(), self.flag_string())

  def print_debug_info(self, read):
    print >>sys.stderr, "bwa_seq_t for read"
    print >>sys.stderr, 'name:', read.get_name()
    print >>sys.stderr, 'seq:', ''.join(['ACGTN'[read.seq[k]] for k in xrange(read.full_len)])
    print >>sys.stderr, 'rseq:', ''.join(['ACGTN'[read.rseq[k]] for k in xrange(read.full_len)])
    if read.qual:
      print >>sys.stderr, 'qual:', ''.join([chr(read.qual[k]) for k in xrange(read.full_len)])
    else:
      print >>sys.stderr, 'qual: NULL'
    print >>sys.stderr, 'len:', read.len
    print >>sys.stderr, 'strand:', read.strand
    print >>sys.stderr, 'type:', read.type
    print >>sys.stderr, 'dummy:', read.dummy
    print >>sys.stderr, 'extra_flag: 0x%x' % read.extra_flag
    print >>sys.stderr, 'n_mm:', read.n_mm
    print >>sys.stderr, 'n_gapo:', read.n_gapo
    print >>sys.stderr, 'n_gape:', read.n_gape
    print >>sys.stderr, 'mapQ:', read.mapQ
    print >>sys.stderr, 'score:', read.score
    print >>sys.stderr, 'clip_len:', read.clip_len
    print >>sys.stderr, 'n_aln:', read.n_aln
    print >>sys.stderr, 'aln: --'
    print >>sys.stderr, 'n_multi:', read.n_multi
    print >>sys.stderr, 'multi: --'
    print >>sys.stderr, 'sa:', read.sa
    print >>sys.stderr, 'pos:', read.pos
    print >>sys.stderr, 'c1:', read.c1
    print >>sys.stderr, 'c2:', read.c2
    print >>sys.stderr, 'seQ:', read.seQ
    print >>sys.stderr, 'n_cigar:', read.n_cigar
    print >>sys.stderr, 'cigar: ', [ (read.cigar[i] & CIGAR_LN_MASK, read.cigar[i] >> CIGAR_OP_SHIFT) for i in xrange(read.n_cigar) ]
    print >>sys.stderr, 'tid:', read.tid
    print >>sys.stderr, 'full_len:', read.full_len
    print >>sys.stderr, 'nm:', read.nm
    print >>sys.stderr, 'md: --'



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Seal is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Seal.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import sys

from bl.lib.tools.seal_config_file import SealConfigFile, FormatError

class SeqalConfigError(RuntimeError):
	pass

class SeqalConfig(object):
	"""
	Reads the command line and the Seal config file.
	Merges options from both sources, giving priority to the command line.
	"""

	ConfigSection = "Seqal"

	class Args(argparse.Namespace):
		def __init__(self):
			self.properties = {}

	class SetProperty(argparse.Action):
		"""
		Used with argparse to parse arguments setting property values.
		Creates an attribute 'property' in the results namespace containing
		all the property-value pairs read from the command line.
		"""
		def __call__(self, parser, namespace, value, option_string=None):
			name, v = value.split('=', 1)
			namespace.properties[name] = v

	class SetTrimQProperty(argparse.Action):
		"""
		When handling the --trimq option, instead of setting a trimq attribute on the
		namespace object we use this action to directly set the correct seqal config property.
		"""
		def __call__(self, parser, namespace, value, option_string=None):
			namespace.properties['seal.seqal.trim.qual'] = value

	def __init__(self):
		self.cmd_parser = argparse.ArgumentParser(description='Distributed BWA read alignment and duplicates removal.')
		# make the parser print help whenever there's a usage error
		def error(message):
			sys.stderr.write('error: %s\n\n' % message)
			self.cmd_parser.print_help()
			sys.exit(2)

		self.cmd_parser.error = error
		##############

		self.cmd_parser.add_argument('input', metavar='INPUT', help='input path')
		self.cmd_parser.add_argument('output', metavar='OUTPUT', help='output path')
		self.cmd_parser.add_argument('reference', metavar='REF.tar', help='reference archive (tar or tar.gz)')
		self.cmd_parser.add_argument('-q', '--trimq', metavar='Q', type=int, action=type(self).SetTrimQProperty,
				help="trim quality, like BWA's -q argument (default: 0).")
		self.cmd_parser.add_argument('-a', '--align-only', action='store_true',
				help="Only perform alignmnet and skip duplicates detection (default: false).")
		self.cmd_parser.add_argument('-r', '--num-reducers', metavar='INT', type=int, dest="num_reducers",
				help="Number of reduce tasks. Specify 0 to perform alignment without duplicates removal (default: 3 * num task trackers).")
		self.cmd_parser.add_argument('-sc', '--seal-config', metavar='FILE', dest="seal_config", default=os.path.join(os.path.expanduser('~'), '.sealrc'),
				help='Override the default Seal config file')
		self.cmd_parser.add_argument('-D', metavar="PROP=VALUE", action=type(self).SetProperty,
				help='Set a property value, such as -D mapred.compress.map.output=true')

	def load_config_and_cmd_line(self, argv=sys.argv[1:]):
		# we scan the command line first in case the user wants to
		# override the default config file location
		args, left_over = self.cmd_parser.parse_known_args(args=argv, namespace=SeqalConfig.Args())

		# load the config for this program, if the file exists
		config = SealConfigFile()

		# was a config file different from the default specified on the command line?
		try:
			if args.seal_config != self.cmd_parser.get_default("seal_config"):
				# in this case, make sure it exists and is readable
				if not os.path.exists(args.seal_config):
					raise SeqalConfigError("The specified Seal config file %s doens't exist" % args.seal_config)
				if not os.access(args.seal_config, os.R_OK):
					raise SeqalConfigError("The specified Seal config file %s isn't readable" % args.seal_config)
				config.read(args.seal_config) # no problems.  Load the file.
			else:
				# check the default path.  If the file exists and is readable we'll load it
				if os.path.exists(args.seal_config):
					if os.access(args.seal_config, os.R_OK):
						config.read(args.seal_config)
					else:
						print >>sys.stderr, "WARNING:  Seal config file %s exists but isn't readable" % args.seal_config
		except FormatError as e: # catch errors from parsing the config file
			raise SeqalConfigError("Error in Seal configuration file %s\n%s" % (args.seal_config, str(e)))

		# override configuration properties from file with the ones
		# provided on the command line.
		for name, value in config.items(SeqalConfig.ConfigSection):
			if not args.properties.has_key(name):
				args.properties[name] = value

		return args, left_over



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Seal is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Seal.  If not, see <http://www.gnu.org/licenses/>.


import ConfigParser
import re

# This class provides functionality very similar to ConfigParser.
# There are a few differences however:
# * when a requested section doesn't exist ConfigParser raises an error
#   (NoSectionError) while SealConfigFile shows the default values.
# * case sensitive section and key names (except DEFAULT)
# * ignore blanks at the start of a line
# * when getting an option that doesn't exist SealConfigFile returns None (ConfigParser raises NoOptionError)

class FormatError(RuntimeError):
	pass

class SealConfigFile(object):

	DefaultSectionName = "DEFAULT"

	def __init__(self):
		self.__config = self.__make_config()

	def __make_config(self):
		# make a config object, but overwrite is key formatting function to
		# make it case sensitive and trim the spaces around the key.
		cfg = ConfigParser.SafeConfigParser()
		cfg.optionxform = lambda option: option.strip()
		return cfg

	def read(self, filename):
		with open(filename) as f:
			self.readfp(f)

	def readfp(self, fp):
		# A custom reader for the config file since we slightly modified the
		# Python ConfigFile format:
		# * case sensitive section and key names (except DEFAULT)
		# * ignore blanks at the start of a line
		self.__config = self.__make_config()
		re_section = re.compile('^\[ *(\w+) *\] *$')
		re_kv_line = re.compile('^ *([^:=]+)[:=] *(.*) *$')
		current_section = None
		lineno = 0

		for line in fp:
			lineno += 1
			line = line.strip()
			if line == "" or line[0] == '#' or line[0] == ';':
				# comment or blank line
				continue
			else:
				m_section = re_section.match(line)
				m_kv = re_kv_line.match(line)
				if m_section:
					current_section = m_section.group(1)
					if current_section != self.__class__.DefaultSectionName:
						self.__config.add_section(current_section)
				elif m_kv:
					k,v = m_kv.group(1,2)
					if current_section is None or current_section.upper() == self.__class__.DefaultSectionName:
						self.__config.set("DEFAULT", k, v)
					else:
						self.__config.set(current_section, k, v)
				else:
					raise FormatError("Invalid config file line %d: %s" % (lineno, line))


	def has_section(self, name):
		"""Whether the config has the named section.  Unlike ConfigParser, returns true for section "default"."""
		if name is not None and name.upper() == "DEFAULT":
			return True
		else:
			return self.__config.has_section(name)

	def sections(self):
		return self.__config.sections()

	def get(self, section, key):
		"""Unlike ConfigParser, if the section doesn't exist is looks
		for the key among the defaults (ConfigParser raises NoSectionError).
		Returns None if the section-key doesn't exist."""
		try:
			if self.__config.has_section(section):
				return self.__config.get(section, key)
			else:
				return self.__config.get("DEFAULT", key)
		except ConfigParser.NoOptionError:
			return None

	def items(self, section):
		if self.__config.has_section(section):
			return self.__config.items(section)
		else:
			return self.__config.items("DEFAULT")



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import os
import ctypes as ct
from constants import *

libbwa_path = os.path.join(os.path.split(__file__)[0], 'libbwa.so')
libbwa = ct.CDLL(libbwa_path)

Q_OFFSET = {
  "fastq-sanger": 33,
  "fastq-solexa": 64,
  "fastq-illumina": 64
}

################
# Basic types
################
p_char_p = ct.POINTER(ct.c_char_p)
################

def make_cigar(raw_cigar, n_cigar, read_len):
  if n_cigar == 0:
    return [(read_len, 'M')]
  cigar = []
  for k in xrange(n_cigar):
    size = raw_cigar[k] & CIGAR_LN_MASK
    op = "MIDS"[raw_cigar[k]>>CIGAR_OP_SHIFT]
    cigar.append((size, op))
  return cigar


def get_cigar_pos_end(cigar, pos):
  """Returns the position one beyond the last coordinate of the cigar"""
  x = pos
  for size, op in cigar:
    if op in 'MD':
      x += size
  return x

#-STRUCT-WRAPPERS--------------------------------------------------------------
class gap_opt_t(ct.Structure):
  _fields_ = [("s_mm", ct.c_int32),
              ("s_gapo", ct.c_int32),
              ("s_gape", ct.c_int32),
              ("mode", ct.c_int32),
              ("indel_end_skip", ct.c_int32),
              ("max_del_occ", ct.c_int32),
              ("max_entries", ct.c_int32),
              ("fnr", ct.c_float),
              ("max_diff", ct.c_int32),
              ("max_gapo", ct.c_int32),
              ("max_gape", ct.c_int32),
              ("max_seed_diff", ct.c_int32),
              ("seed_len", ct.c_int32),
              ("n_threads", ct.c_int32),
              ("max_top2", ct.c_int32),
              ("trim_qual", ct.c_int32),
              ]


class pe_opt_t(ct.Structure):
  _fields_ = [("max_isize", ct.c_int),
              ("max_occ", ct.c_int),
              ("n_multi", ct.c_int),
              ("N_multi", ct.c_int),
              ("type", ct.c_int),
              ("is_sw", ct.c_int),
              ("is_preload", ct.c_int),
              ("ap_prior", ct.c_double)]


class bwt_t(ct.Structure):
  _fields_ = [("primary", ct.c_uint32), #S^{-1}(0), or the primary index of BWT
              ("L2", ct.c_uint32 * 5), #C(), cumulative count
              ("seq_len", ct.c_uint32), # sequence length
              ("bwt_size", ct.c_uint32), # size of bwt, about seq_len/4
              ("bwt", ct.POINTER(ct.c_uint32)), # BWT
              ("cnt_table", ct.c_uint32 * 256),  # cnt_table[256]
              ("sa_intv", ct.c_int32),
              ("n_sa", ct.c_uint32),
              ("sa", ct.POINTER(ct.c_uint32))]


class bntann1_t(ct.Structure):
  _fields_ = [("offset", ct.c_int64),
              ("len", ct.c_int32),
              ("n_ambs", ct.c_int32),
              ("gi", ct.c_uint32),
              ("name", ct.c_char_p),
              ("anno", ct.c_char_p)]


class bntamb1_t(ct.Structure):
  _fields_ = [("offset", ct.c_int64),
              ("len", ct.c_int32),
              ("amb", ct.c_char)]


class bntseq_t(ct.Structure):
  _fields_ = [("l_pac", ct.c_int64),
              ("n_seqs", ct.c_int32),
              ("seed", ct.c_uint32),
              ("anns", ct.POINTER(bntann1_t)), # n_seqs elements
              ("n_holes", ct.c_int32),
              ("ambs", ct.POINTER(bntamb1_t)), # n_holes elements
              ("fp_pac", ct.c_void_p)] # this is really a FILE *fp_pac

class bwt_width_t(ct.Structure):
  _fields_ = [('w', ct.c_uint32),
              ('bid', ct.c_int32)]


class bwt_aln1_t(ct.Structure):
  _fields_ = [('n_mm', ct.c_uint32, 8),
              ('n_gapo', ct.c_uint32, 8),
              ('n_gape', ct.c_uint32, 8),
              ('a', ct.c_uint32, 1),
              ('k', ct.c_uint32),
              ('l', ct.c_uint32),
              ('score', ct.c_int32)]

# CRS4 patch: 'gap' split into 'n_gapo' + 'n_gape'; added score; 'mm' => 'n_mm'
class bwt_multi1_t(ct.Structure):
  _fields_ = [('pos', ct.c_uint32),
              ('n_cigar', ct.c_uint32, 15),
              ('n_gapo', ct.c_uint32, 8),
              ('n_gape', ct.c_uint32, 8),
              ('n_mm', ct.c_uint32, 8),
              ('strand', ct.c_uint32, 1),
              ('score', ct.c_int32),
              ('cigar', ct.POINTER(ct.c_uint16))]

  def __init__(self):
    super(bwt_multi1_t, self).__init__()

  def get_cigar(self, read_len):
    if not hasattr(self, "__cigar"):
      self.__cigar = make_cigar(self.cigar, self.n_cigar, read_len)
    return self.__cigar

  def get_pos_end(self, read_len):
    if not hasattr(self, "__pos_end"):
      self.__pos_end = get_cigar_pos_end(self.get_cigar(read_len), self.pos)
    return self.__pos_end


class bwa_seq_t(ct.Structure):
  _fields_ = [('name', ct.c_char_p),
              ('seq', ct.POINTER(ct.c_uint8)),
              ('rseq', ct.POINTER(ct.c_uint8)),
              ('qual', ct.POINTER(ct.c_uint8)),
              ('len', ct.c_uint32, 20),
              ('strand', ct.c_uint32, 1),
              ('type', ct.c_uint32, 2),
              ('dummy', ct.c_uint32, 1),
              ('extra_flag', ct.c_uint32, 8),
              ('n_mm', ct.c_uint32, 8),
              ('n_gapo', ct.c_uint32, 8),
              ('n_gape', ct.c_uint32, 8),
              ('mapQ', ct.c_uint32, 8),
              ('score', ct.c_int32),
              ('clip_len', ct.c_int32),
              ('n_aln', ct.c_int32),
              ('aln', ct.POINTER(bwt_aln1_t)),
              ('n_multi', ct.c_int32),
              ('multi', ct.POINTER(bwt_multi1_t)),
              ('sa', ct.c_uint32),
              ('pos', ct.c_uint32),
              ('c1', ct.c_uint64, 28),
              ('c2', ct.c_uint64, 28),
              ('seQ', ct.c_uint64, 8),
              ('n_cigar', ct.c_int32),
              ('cigar', ct.POINTER(ct.c_uint16)),
              ('tid', ct.c_int32),
							('bc', ct.c_char * 16), # in-struct array, up to 15 bases, null-terminated
              ('full_len', ct.c_uint32, 20),
              ('nm', ct.c_uint32, 12),
              ('md', ct.c_char_p)]

  def __len__(self):
    """
    Return the full length of the sequence.  Note that only
    the section up to clip_len is used for the alignment.  Read
    trimming in fact modifies the value of clip_len."""
    return self.full_len

  def __compute_cigar_ops(self):
    """
    Returns an array of tuples (lenth, op) corresponding to
    the alignment of this sequence.

    Based on bwa_print_sam1 from bwase.c
    """
    if self.type == BWA_TYPE_NO_MATCH:
      return []
    elif self.n_cigar == 0:
      return [(self.len, 'M')]
    else:
      cigar = []
      for k in xrange(self.n_cigar):
        size = self.cigar[k] & CIGAR_LN_MASK
        op = "MIDS"[self.cigar[k]>>CIGAR_OP_SHIFT]
        cigar.append((size, op))
      return cigar

  def __compute_pos_end(self):
    """
    Computes the position on the reference where this aligned
    sequence ends, based on the alignment described by the
    CIGAR operations and the starting position self.pos.
    Based on pos_end in bwase.c.
    NOTE:  pos_end uses p->len instead of full_len or clip_len.
    Raises: ValueError if self is of type BWA_TYPE_NO_MATCH
    """
    if self.type == BWA_TYPE_NO_MATCH:
      raise ValueError("Cannot calculate end position of an unmapped sequence")
    cig = self.get_cigar()
    if len(cig) == 0:
      return self.pos + self.len
    else:
      return get_cigar_pos_end(cig, self.pos)

  def get_cigar(self):
    if not hasattr(self, "__cigar"):
      self.__cigar = self.__compute_cigar_ops()
    return self.__cigar

  def get_pos_end(self):
    if not hasattr(self, "__pos_end"):
      self.__pos_end = self.__compute_pos_end()
    return self.__pos_end

  def get_pos_5(self):
    """
    position of the 5' end.
    Raises: ValueError if self is of type BWA_TYPE_NO_MATCH
    """
    if self.type == BWA_TYPE_NO_MATCH:
      raise ValueError("Cannot calculate 5' position of an unmapped sequence")
    if self.strand:
      return self.get_pos_end()
    else:
      return self.pos

  def get_seq(self):
    """
    Return seq (the original sequence) as a text string.
    """
    if not hasattr(self, "__seq"):
      self.__seq = ''.join(['ACGTN'[self.seq[k]] for k in xrange(self.full_len)])
    return self.__seq

  def get_rseq(self):
    """
    Return complement of seq as a text string.
    """
    # XXX:  note that for in BWA rseq is practically a temporary buffer.
    # It's only correct in the range [0,clip_len).  BWA itself creates the
    # complement sequence for output directly from the seq array.
    if not hasattr(self, "__rseq"):
      self.__rseq = ''.join(['TGCAN'[self.seq[k]] for k in xrange(self.full_len-1, -1, -1)])
    return self.__rseq

  # Get a string with the quality of the bases using an ASCII-33 encoding
  def get_qual_string(self):
    if not hasattr(self, "__qual"):
      ptr = self.qual
      if not ptr:
        self.__qual = ''
      else:
        self.__qual = ''.join([chr(ptr[k]) for k in xrange(self.full_len)])
    return self.__qual

  def get_name(self):
    return self.name

  def itermulti(self):
    """
    Yields alignment-specific info for all hits, including the 'main' one.
    """
    yield self
    for i in xrange(self.n_multi):
      yield self.multi[i]


bwa_seq_p_t = ct.POINTER(bwa_seq_t)


class isize_info_t(ct.Structure):
  _fields_ = [('avg', ct.c_double),
              ('std', ct.c_double),
              ('low', ct.c_uint32),
              ('high', ct.c_uint32),
              ('high_bayesian', ct.c_uint32)]


#-END-STRUCT-WRAPPERS----------------------------------------------------------
init_g_log_n = libbwa.init_g_log_n
init_g_log_n.argtypes = []
init_g_log_n()

init_g_hash = libbwa.init_g_hash
init_g_hash.argtypes = []
init_g_hash()

#------------------------------------------------------------
gap_opt_p_t = ct.POINTER(gap_opt_t)
gap_init_opt = libbwa.gap_init_opt
gap_init_opt.argtypes = []
gap_init_opt.restype = gap_opt_p_t

#------------------------------------------------------------
pe_opt_p_t = ct.POINTER(pe_opt_t)
pe_init_opt = libbwa.bwa_init_pe_opt
pe_init_opt.argtypes = []
pe_init_opt.restype = pe_opt_p_t

#------------------------------------------------------------
bntseq_p_t = ct.POINTER(bntseq_t)

bns_restore = libbwa.bns_restore
bns_restore.argtypes = [ct.c_char_p]
bns_restore.restype = bntseq_p_t
restore_bns=bns_restore

bns_destroy_base = libbwa.bns_destroy
bns_destroy_base.argtypes = [bntseq_p_t]

#------------------------------------------------------------
bwt_p_t = ct.POINTER(bwt_t)
bwt_restore_bwt = libbwa.bwt_restore_bwt
bwt_restore_bwt.argtypes = [ct.c_char_p]
bwt_restore_bwt.restype = bwt_p_t
restore_bwt=bwt_restore_bwt

bwt_restore_sa = libbwa.bwt_restore_sa
bwt_restore_sa.argtypes = [ct.c_char_p, bwt_p_t]
restore_sa=bwt_restore_sa

bwt_destroy_base = libbwa.bwt_destroy
bwt_destroy_base.argtypes = [bwt_p_t]

#--
bwt_restore_bwt_mmap = libbwa.bwt_restore_bwt_mmap
bwt_restore_bwt_mmap.argtypes = [ct.c_char_p]
bwt_restore_bwt_mmap.restype = bwt_p_t
restore_bwt_mmap=bwt_restore_bwt_mmap

bwt_restore_sa_mmap = libbwa.bwt_restore_sa_mmap
bwt_restore_sa_mmap.argtypes = [ct.c_char_p, bwt_p_t]
restore_sa_mmap=bwt_restore_sa_mmap

bwt_destroy_mmap = libbwa.bwt_destroy_mmap
bwt_destroy_mmap.argtypes = [bwt_p_t]

ubyte_p_t = ct.POINTER(ct.c_ubyte)
bwt_restore_reference_mmap_helper = libbwa.bwt_restore_reference_mmap_helper
bwt_restore_reference_mmap_helper.argtypes = [ct.c_char_p, ct.c_int32]
bwt_restore_reference_mmap_helper.restype  = ubyte_p_t


#------------------------------------------------------------
bwa_cal_sa_reg_gap_mt = libbwa.bwa_cal_sa_reg_gap_mt
#    (bwt_t *const bwt[2], int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt) ;
bwa_cal_sa_reg_gap_mt.argtypes = [bwt_p_t * 2, ct.c_int32, bwa_seq_p_t, gap_opt_p_t]
cal_sa_reg_gap_mt = bwa_cal_sa_reg_gap_mt

#------------------------------------------------------------
bwa_cal_sa_reg_gap = libbwa.bwa_cal_sa_reg_gap
#    (int tid, bwt_t *const bwt[2], int n_seqs, bwa_seq_t *seqs, const gap_opt_t *opt) ;
bwa_cal_sa_reg_gap.argtypes = [ct.c_int32, bwt_p_t * 2, ct.c_int32, bwa_seq_p_t, gap_opt_p_t]
cal_sa_reg_gap = bwa_cal_sa_reg_gap


#------------------------------------------------------------
bwa_cal_pac_pos_pe = libbwa.bwa_cal_pac_pos_pe_memory
bwa_cal_pac_pos_pe.argtypes = [bwt_p_t * 2, ct.c_int, bwa_seq_p_t * 2,
                               ct.POINTER(isize_info_t),
                               pe_opt_p_t, gap_opt_p_t, ct.POINTER(isize_info_t)]
bwa_cal_pac_pos_pe.restype  = ct.c_int
#------------------------------------------------------------
bns_coor_pac2real = libbwa.bns_coor_pac2real
#int bns_coor_pac2real(const bntseq_t *bns, int64_t pac_coor, int len, int32_t *real_seq)
bns_coor_pac2real.argtypes = [bntseq_p_t, ct.c_uint64, ct.c_uint,
                               ct.POINTER(ct.c_uint32)]
bns_coor_pac2real.restype = ct.c_int

#------------------------------------------------------------
bwa_trim_read = libbwa.bwa_trim_read
# int bwa_trim_read(int trim_qual, bwa_seq_t *p)
bwa_trim_read.argtypes = [ct.c_int, bwa_seq_p_t]
bwa_trim_read.restype = ct.c_int

#------------------------------------------------------------
bwa_seq_reverse = libbwa.seq_reverse
# void seq_reverse(int len, ubyte_t *seq, int is_comp)
bwa_seq_reverse.argtypes = [ ct.c_int, ct.POINTER(ct.c_ubyte), ct.c_int ]
bwa_seq_reverse.restype = None

#------------------------------------------------------------
init_sequences = libbwa.bwa_init_sequences
# bwa_seq_t* bwa_init_sequences(const char**const data2d, const size_t n_seqs, const int q_offset, const int trim_qual)
init_sequences.argtypes = [p_char_p, ct.c_uint, ct.c_int, ct.c_int]
init_sequences.restype = bwa_seq_p_t

#------------------------------------------------------------
# Start Python functions

def get_seq_id(bns, pac_coordinate, ref_len):
  seq_id = ct.c_uint32()
  nn = bns_coor_pac2real(bns, pac_coordinate, ref_len, ct.byref(seq_id))
  return seq_id.value, nn


def cal_pac_pos_pe(bwts, n_seqs, bwsa, ii, popt, gopt, last_ii):
  bwsa_c = (bwa_seq_p_t * 2)()
  bwsa_c[0] = bwsa[0]
  bwsa_c[1] = bwsa[1]
  res = bwa_cal_pac_pos_pe(bwts, n_seqs, bwsa_c, ct.byref(ii), popt, gopt,
                           ct.byref(last_ii))
  return res

#------------------------------------------------------------
#ubyte_t *bwa_paired_sw(const bntseq_t *bns, const ubyte_t *_pacseq,
#                       int n_seqs, bwa_seq_t *seqs[2],
#                       const pe_opt_t *popt, const isize_info_t *ii);

bwa_paired_sw = libbwa.bwa_paired_sw
bwa_paired_sw.argtypes = [bntseq_p_t, ct.POINTER(ct.c_uint8), ct.c_int,
                          bwa_seq_p_t * 2,
                          pe_opt_p_t,
                          ct.POINTER(isize_info_t)]
bwa_paired_sw.restype = ct.POINTER(ct.c_uint8)

def paired_sw(bns, pacseq, n_seq_pairs, bwsa, popt, ii, offset=0):
  bwsa_c = (bwa_seq_p_t * 2)()
  bwsa_c[0] = ct.cast(ct.byref(bwsa[0][offset]), bwa_seq_p_t)
  bwsa_c[1] = ct.cast(ct.byref(bwsa[1][offset]), bwa_seq_p_t)
  r = bwa_paired_sw(bns, pacseq, n_seq_pairs, bwsa_c, popt, ct.byref(ii))
  return r

#------------------------------------------------------------
bwa_refine_gapped = libbwa.bwa_refine_gapped_memory
bwa_refine_gapped.argtypes = [bntseq_p_t, ct.POINTER(ct.c_uint8), ct.c_int, bwa_seq_p_t]

def refine_gapped(bns, n_seqs, seqs, pacseq):
  seqs_c = ct.cast(seqs, bwa_seq_p_t)
  return bwa_refine_gapped(bns, pacseq,  n_seqs, seqs_c)

#------------------------------------------------------------
bwa_free_read_seq = libbwa.bwa_free_read_seq
bwa_free_read_seq.argtypes = [ct.c_int32, bwa_seq_p_t]
free_seq = bwa_free_read_seq


def build_bws_array(sequence_pairs, qtype="fastq-sanger", trim_qual=0):
  """
  Build an array of BWA seq pairs from a list of sequence pairs.
  A sequence pair is  a (name, read_seq, read_qual, mate_seq, mate_qual).
  """
  # allocate two arrays of bwa_seq_t, one for each set of reads (1 and 2)
  bwsa_t = bwa_seq_p_t * 2
  bwsa = bwsa_t()
  if qtype == "fastq-sanger" or qtype == "fastq-illumina":
    # this is the offset that has been added to the quality scores,
    # which varies depending on the format.
    q_offset = Q_OFFSET[qtype]
  else:
    # XXX: to support fastq-solexa we'd have to rewrite the quality arrays
    # here according to the formula round(q[i] + 10 * log10(1+10**(-q[i]/10.)))
    raise ValueError("sorry. '%s' is not a supported qtype" % qtype)

  row_type = (ct.c_char_p * 3)
  pointers = (row_type*len(sequence_pairs))()
  for read_no in 0,1:
    if read_no == 0:
      row_indices = (0, 1, 2)
    else:
      row_indices = (0, 3, 4)

    for i, row in enumerate(sequence_pairs):
      if len(row) != 5:
        raise ValueError("Invalid row %d.  Expecting (name, read1, qual1, read2, qual2)"
                         " but got a row with length %d\n%s" % (i, len(row), str(row)))
      for j in xrange(3):
        pointers[i][j] = row[ row_indices[j] ]
    bwsa[read_no] = init_sequences( ct.cast(pointers, p_char_p), len(sequence_pairs), q_offset, trim_qual)
    if bwsa[read_no] is None:
      raise RuntimeError("There was an error creating the BWA sequences")
  return bwsa

def restore_index_base(root_name):
  bwts_t = bwt_p_t * 2
  bwts = bwts_t()
  bwts[0] = restore_bwt(root_name + '.bwt')
  bwts[1] = restore_bwt(root_name + '.rbwt')
  restore_sa(root_name + '.sa', bwts[0])
  restore_sa(root_name + '.rsa', bwts[1])
  return bwts


def restore_reference_base(root_name):
  bns = restore_bns(root_name)
  fd = open(root_name + '.pac')
  pacseq_str = fd.read(bns[0].l_pac/4 + 1)
  p = ct.c_char_p(pacseq_str)
  ubyte_p_t = ct.POINTER(ct.c_ubyte)
  pacseq = ct.cast(p, ubyte_p_t)
  return (bns, pacseq)

#--------------------------------------------------------------------


def mmap_magic():
  return ct.c_uint32.in_dll(libbwa, "BWA_MMAP_SA_MAGIC")

import struct
def make_suffix_array_for_mmap(src, dest):
  s = open(src)
  d = open(dest, 'w')
  # Write the magic number in native byte order so that it will only
  # be read correctly by machines that use the same byte ordering
  # (since the data is also in native byte order).
  d.write(struct.pack("=I", mmap_magic().value ))
  d.write(s.read(4*7))
  d.write(struct.pack("=I", 0x0ffffffff)) # uint32_t(-1)
  buf_size = 1000000
  while True:
    c = s.read(buf_size)
    if not c:
      break
    d.write(c)
  d.close()
  s.close()

def make_suffix_arrays_for_mmap(root_name):
  make_suffix_array_for_mmap(root_name + '.sa', root_name + '.sax')
  make_suffix_array_for_mmap(root_name + '.rsa', root_name + '.rsax')

def restore_index_mmap(root_name):
  bwts_t = bwt_p_t * 2
  bwts = bwts_t()
  bwts[0] = restore_bwt_mmap(root_name + '.bwt')
  bwts[1] = restore_bwt_mmap(root_name + '.rbwt')
  restore_sa_mmap(root_name + '.sax', bwts[0])
  restore_sa_mmap(root_name + '.rsax', bwts[1])
  return bwts

def restore_reference_mmap(root_name):
  bns = restore_bns(root_name)
  pac_fname = root_name + '.pac'
  pacseq = bwt_restore_reference_mmap_helper(pac_fname, bns[0].l_pac/4 + 1)
  return (bns, pacseq)

#------------------------------------------------------------

def restore_index(root, mmap=False):
  if mmap:
    return restore_index_mmap(root)
  else:
    return restore_index_base(root)

def restore_reference(root, mmap=False):
  if mmap:
    return restore_reference_mmap(root)
  else:
    return restore_reference_base(root)

def bwt_destroy(bwt, mmap=False):
  if mmap:
    bwt_destroy_mmap(bwt)
  else:
    bwt_destroy_base(bwt)


def bns_destroy(bnsp, mmap=False):
  bns_destroy_base(bnsp)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

SAM_FPD = 1     # paired
SAM_FPP = 2     # properly paired
SAM_FSU = 4     # self-unmapped
SAM_FMU = 8     # mate-unmapped
SAM_FSR = 16    # self on the reverse strand
SAM_FMR = 32    # mate on the reverse strand
SAM_FR1 = 64    # this is read one
SAM_FR2 = 128   # this is read two
SAM_FSC = 256   # secondary alignment
SAM_FQC = 0x200   # failed quality checks
SAM_FDP = 0x400   # PCR or optical duplicate



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

"""
How to use this test:

You need to provide as command line arguments:
  1. The fasta reference sequence file name
  2, 3. The two reads and mates sequence files

The test will print an error count and generate a dump of errors.

See README.test_bwa_big for expected results on known test cases.
"""

import sys, unittest, time, logging, optparse
import testing_utilities as u

import Bio.SeqIO

import itertools as it


class bwa_big_tc(unittest.TestCase):

  PARAMS = None

  @classmethod
  def set_params(cls, args, opt):
    cls.PARAMS = (args, opt)

  def setUp(self):
    (self.refseq_fn, self.read_fn, self.mate_fn), self.opt = self.PARAMS

  def tearDown(self):
    if not self.opt.no_clean:
      for fn in [self.refseq_fn, self.read_fn, self.mate_fn]:
        u.clean_up_aux_files(fn)

  def runTest(self):
    self.__compare_step_by_step()

  def __compare_step_by_step(self):
    if not self.opt.no_index:
      u.build_index(self.refseq_fn)
    #--
    bwa_res = u.run_bwa_sampe(self.refseq_fn, self.read_fn, self.mate_fn)
    py_res =  u.run_bwa_py_sampe(self.refseq_fn, self.read_fn, self.mate_fn,
                                 log_level=self.opt.log_level,
                                 pairing_batch_size=self.opt.pairing_batch_size,
                                 seq_list_len=self.opt.seq_list_len,
                                 fastq_subfmt=self.opt.fastq_subfmt)
    f = open("errors.dump", 'w')
    error_count = 0
    try:
      self.assertEqual(len(bwa_res), len(py_res))
    except AssertionError:
      error_count += 1
      f.write("Error on result n. %d\n" % i)
      f.write("NUMBER OF RESULT FIELDS: %d, %d\n" % (
        len(bwa_res), len(py_res)))
    for i, (o, n) in enumerate(it.izip(bwa_res, py_res)):
      try:
        for k in o.keys():
          if k == 'name':
            nn = n[k].rsplit("/", 1)[0] # we are dropping the '/[12]' extension
            self.assertEqual(o[k], nn)
          elif k == 'aux':
            self.assertEqual(len(o[k]), len(n[k]))
            for x,y in it.izip(o[k], n[k]):
              self.assertEqual(x, y)
          else:
            self.assertEqual(o[k],n[k])
      except AssertionError:
        error_count += 1
        f.write("Error on result n. %d\n" % i)
        for k in o.keys():
          f.write("%r %r %r %r\n" % (k, o[k], n[k], o[k]==n[k]))
        f.write("\n")
    print "error count: %d" % error_count
    f.close()


def suite():
  suite = unittest.TestSuite()
  suite.addTest(bwa_big_tc('runTest'))
  return suite


def make_parser():
    parser = optparse.OptionParser('%prog [OPTIONS] REFSEQ_FN READ_FN MATE_FN')
    parser.add_option(
      "-f", type="str", dest="fastq_subfmt", metavar="STRING",
      default="fastq-illumina", help="fastq subformat (%default)",
      )
    parser.add_option(
      "-l", type=int, dest="seq_list_len", metavar="INT",
      default=10000, help="read this number of seqs at a time (%default)",
      )
    parser.add_option(
      "-p", type=int, dest="pairing_batch_size", metavar="INT", default=None,
      help="seq batch size for paired_sw (None, i.e., do not split in batches)",
      )
    parser.add_option(
      "--no-index", action="store_true", help="do not build build bwa index",
      )
    parser.add_option(
      "--no-clean", action="store_true", help="do not clean up aux files",
      )
    parser.add_option(
      "--log-level", type="choice", dest="log_level", metavar="STRING",
      default="INFO", help="logging level (%default)",
      choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
      )
    return parser


if __name__ == '__main__':

  parser = make_parser()
  opt, args = parser.parse_args(sys.argv)
  opt.log_level = getattr(logging, opt.log_level)
  if len(args) < 4:
    parser.print_help()
    sys.exit(2)
  bwa_big_tc.set_params(args[1:4], opt)

  runner = unittest.TextTestRunner(verbosity=2)
  runner.run((suite()))



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

"""
bntseq.c
--------

void bns_dump(const bntseq_t *bns, const char *prefix)
{
  char str[1024];
  FILE *fp;
  int i;
  { // dump .ann
    strcpy(str, prefix); strcat(str, ".ann");
    fp = xopen(str, "w");
    fprintf(fp, "%lld %d %u\n", (long long)bns->l_pac, bns->n_seqs, bns->seed);
    for (i = 0; i != bns->n_seqs; ++i) {
      bntann1_t *p = bns->anns + i;
      fprintf(fp, "%d %s", p->gi, p->name);
      if (p->anno[0]) fprintf(fp, " %s\n", p->anno);
      else fprintf(fp, "\n");
      fprintf(fp, "%lld %d %d\n", (long long)p->offset, p->len, p->n_ambs);
    }
    fclose(fp);
  }
"""

import sys
import bl.lib.seq.aligner.bwa as bwa

try:
  ROOT = sys.argv[1]
except IndexError:
  sys.exit("USAGE: %s ROOT_REF_NAME (e.g., refseq.fa)")

bnsp, pacseq = bwa.restore_reference(ROOT)


#regenerate .ann file
print bnsp[0].l_pac, bnsp[0].n_seqs, bnsp[0].seed
for i in xrange(bnsp[0].n_seqs):
  p = bnsp[0].anns[i]
  print p.gi, p.name, (p.anno if p.anno else "")
  print p.offset, p.len, p.n_ambs



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

from event_monitor import EventMonitor
import time

class StandardMonitor(EventMonitor):
	"""
	Implementation of EventMonitor that writes to a logger
	and implements counting and timing internally.
	"""
	def __init__(self, logger):
		self.__output = logger
		self.__counters = {}
		self.__start_times = {}

	def start(self, event_name):
		self.__start_times[event_name] = time.time()

	def stop(self, event_name):
		delta = time.time() - self.__start_times[event_name]
		self.__output.info( "done with %s (%.3f s)" % (event_name, delta))

	def count(self, event_name, value=1):
		if not self.__counters.has_key(event_name):
			self.__counters[event_name] = value
		else:
			self.__counters[event_name] += value

	def has_counter(self, event_name):
		self.__counters.has_key(event_name)

	def add_counter(self, event_name, display_name=None):
		if self.__counters.has_key(name):
			raise ValueError("counter '%s' is already defined" % name)
		self.__counters[event_name] = 0

	def new_status(self, status_msg):
		self.__output.info(status_msg)

	def log_debug(self, *args):
		self.__output.debug(*args)

	def log_info(self, *args):
		self.__output.info(*args)

	def log_warning(self, *args):
		self.__output.warning(*args)

	def log_error(self, *args):
		self.__output.error(*args)

	def log_critical(self, *args):
		self.__output.critical(*args)

	def each_counter(self):
		for k,v in self.__counters.items():
			yield k,v



#!/usr/bin/env python

# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Seal is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Seal.  If not, see <http://www.gnu.org/licenses/>.

#########################################################
# You need a working Hadoop cluster to use this.
#########################################################


import traceback
import subprocess
import os
import sys
import shutil
import site
SealDir = os.path.realpath( os.path.join( os.path.dirname( os.path.realpath(__file__) ), "..", "..") )
site.addsitedir(SealDir)

import bl.lib.tools.hadut as hadut

class SealIntegrationTest(object):

	def __init__(self, test_dir):
		#super(self.__class__, self).__init__(self)
		self.test_dir = test_dir
		self.test_name = os.path.basename(test_dir)
		self.jar = hadut.find_seal_jar(SealDir)
		self.output_dir = "/tmp/%s.%d" % (self.test_name, os.getpid())

	def setup(self):
		#   Jar: the Seal jar
		#   wd:  working HDFS directory (same as the test name).
		#   OutputDir:  temporary output directory where to store temp files.
		hadut.run_hadoop_cmd_e("dfsadmin", args_list=["-safemode", "wait"])
		self.log("hdfs in safe mode")

		hadut.dfs("-mkdir", self.make_hdfs_test_path())
		input_dir = self.make_hdfs_input_path()
		hadut.dfs("-put", self.make_local_input_path(), input_dir)

	def test_method(self):
		self.log(self.__class__.__name__, "running test")
		self.setup()

		try:
			success = False
			self.log("running program")
			self.run_program(self.make_hdfs_input_path(), self.make_hdfs_output_path())
			self.log("now going to process input")
			self.log("""hadut.dfs("-get", %s, %s)""" % (self.make_hdfs_output_path(), self.output_dir))
			hadut.dfs("-get", self.make_hdfs_output_path(), self.output_dir)
			success = self.process_output()
		finally:
			self.log("cleaning up")
			self.clean_up()

		return success

	def clean_up(self):
		hadut.dfs("-rmr", self.make_hdfs_input_path())
		hadut.dfs("-rmr", self.make_hdfs_output_path())
		hadut.dfs("-rmr", self.make_hdfs_test_path())

	def run_program(hdfs_input, hdfs_output):
		raise NotImplementedError()

	# Compares a file containing the expected test output to
	# the test output, sorted with a plain call to "sort".
	# Shows the diff if there are any.
	#
	# @param expected_file:  file containing expected output
	# @param hadoop_output_dir:  directory where the hadoop part-r-xxxx files are.  If
	#         not specified, this is assumed to be self.output_dir
	# @returns: False if they they are the same, True if they are different
	def sorted_output_different(self, expected_file, hadoop_output_dir=None):
		if hadoop_output_dir is None:
			hadoop_output_dir = self.output_dir
		sorted_output = "%s/sorted_output" % self.output_dir

		retcode = os.system( """cat "%s"/part-* | LC_ALL=C sort > "%s" """ % (hadoop_output_dir, sorted_output) )
		if retcode != 0:
			raise RuntimeError("%d return code when running cat|sort" % retcode)
		print """diff "%s" "%s" """ % (expected_file, sorted_output)
		retcode = os.system("""diff "%s" "%s" """ % (expected_file, sorted_output))
		return retcode != 0

	# print a message with the test result (successful/unsuccessful).
	def show_test_msg(self, successful):
		if successful:
			print "Test successful!"
		else:
			print >>sys.stderr, "Error!  Unexpected result in test %s" % self.test_name

	# Self explanatory.
	def rm_output_dir(self):
		try:
			shutil.rmtree(self.output_dir)
		except OSError as e:
			print >>sys.stderr, "Tried deleting output directory '%s' but got an error (%s)" % (self.output_dir, e)

	# A canned function for generic test output processing.
	# If it doesn't fit your needs use the functions above to piece
	# together the required functionality.
	#
	# Reads the output ${OutputDir}/part-*
	# Compares it with the sorted text in the file "expected"
	# returns True if the output is as expected, else False
	def process_output(self):
		different = self.sorted_output_different( self.make_local_expected_output_path(), self.output_dir )
		self.show_test_msg( not different )
		self.rm_output_dir()
		return not different

	def get_test_dir(self):
		return self.test_dir

	def make_hdfs_test_path(self):
		return os.path.basename(self.test_dir)

	def make_hdfs_input_path(self):
		return os.path.join( self.make_hdfs_test_path(), "input")

	def make_local_input_path(self):
		return os.path.join(self.get_test_dir(), "input")

	def make_hdfs_output_path(self):
		return os.path.join( self.make_hdfs_test_path(), "output")

	def make_local_expected_output_path(self):
		return os.path.join(self.get_test_dir(), "expected")

	def log(self, *args):
		print >>sys.stderr, " ".join(args)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

"""
How to use this test:

You need to provide as command line arguments:
  1. The fasta reference sequence file name
  2, 3. The two reads and mates sequence files

Change SEQ_LIST_LEN_RANGE as needed -- the test measures peak memory
and number of failed inferences of insert size in function of the
length of the sequence list.
"""

import sys, gc, logging
#print gc.isenabled()
gc.set_debug(gc.DEBUG_UNCOLLECTABLE)

import itertools as it
import bl.lib.seq.aligner.bwa as bwa

import Bio.SeqIO
from bl.lib.util.meminfo import meminfo


MB = float(2**20)


def print_meminfo(status, log_function=logging.debug):
  minfo = meminfo()
  size = minfo["size"]/MB
  resident = minfo["resident"]/MB
  msg = "%s: size = %.1fm; resident = %.1fm" % (status, size, resident)
  log_function(msg)
  return size, resident


def run_bwa_py_sampe(refseq_fname, read_fname, mate_fname, seq_list_len=None):
  size_list = []
  resident_list = []
  failed_ii = 0

  read_flow = Bio.SeqIO.parse(open(read_fname), 'fastq-illumina')
  mate_flow = Bio.SeqIO.parse(open(mate_fname), 'fastq-illumina')

  #pairs = [x for x in it.izip(read_flow, mate_flow)]
  pairs_flow = it.izip(read_flow, mate_flow)

  while 1:
    pairs = list(it.islice(pairs_flow, 0, seq_list_len))
    if len(pairs) == 0:
      break
    size, resident = print_meminfo("AFTER READING PAIRS")
    size_list.append(size)
    resident_list.append(resident)

    bwsa = bwa.build_bws_array(pairs)
    size, resident = print_meminfo("AFTER BUILDING BWSA")
    size_list.append(size)
    resident_list.append(resident)

    bwts = bwa.restore_index(refseq_fname)
    size, resident = print_meminfo("AFTER RESTORING INDEX")
    size_list.append(size)
    resident_list.append(resident)

    bnsp, pacseq = bwa.restore_reference(refseq_fname)
    size, resident = print_meminfo("AFTER RESTORING REFERENCE")
    size_list.append(size)
    resident_list.append(resident)

    gopt, popt = bwa.gap_init_opt(), bwa.pe_init_opt()
    ii, last_ii = bwa.isize_info_t(), bwa.isize_info_t()
    last_ii.avg = -1.0
    l = len(pairs)
    size, resident = print_meminfo("AFTER INIT OPT & II")
    size_list.append(size)
    resident_list.append(resident)

    bwa.cal_sa_reg_gap(0, bwts, l, bwsa[0], gopt)
    bwa.cal_sa_reg_gap(0, bwts, l, bwsa[1], gopt)
    size, resident = print_meminfo("AFTER CAL_SA_REG_GAP")
    size_list.append(size)
    resident_list.append(resident)

    cnt_chg = bwa.cal_pac_pos_pe(bwts, l, bwsa, ii, popt, gopt, last_ii)
    size, resident = print_meminfo("AFTER CAL_PAC_POS_PE")
    size_list.append(size)
    resident_list.append(resident)

    #sys.stderr.write("ii=%f\n" % ii.avg)
    if ii.avg < 0.0:
      failed_ii += 1

    bwa.paired_sw(bnsp, pacseq, l, bwsa, popt, ii)
    size, resident = print_meminfo("AFTER PAIRED_SW")
    size_list.append(size)
    resident_list.append(resident)

    bwa.refine_gapped(bnsp, l, bwsa[0], pacseq)
    bwa.refine_gapped(bnsp, l, bwsa[1], pacseq)
    size, resident = print_meminfo("AFTER REFINE_GAPPED")
    size_list.append(size)
    resident_list.append(resident)

    for k in xrange(l):
      v1 = bwa.analyze_hit(gopt[0], bnsp, bwsa[0][k], bwsa[1][k])
      v2 = bwa.analyze_hit(gopt[0], bnsp, bwsa[1][k], bwsa[0][k])
    size, resident = print_meminfo("AFTER ANALYZE_HIT")
    size_list.append(size)
    resident_list.append(resident)

    # deallocate seq & ref data
    for i in 0, 1:
      bwa.free_seq(l, bwsa[i])
      bwa.bwt_destroy(bwts[i])
    bwa.bns_destroy(bnsp)
    size, resident = print_meminfo("AFTER DEALLOC")
    size_list.append(size)
    resident_list.append(resident)

    del pacseq
    n_unreachable = gc.collect()
    logging.debug("n_unreachable = %d" % n_unreachable)
    size, resident = print_meminfo("AFTER DEL PACSEQ")
    size_list.append(size)
    resident_list.append(resident)

    del pairs
    n_unreachable = gc.collect()
    logging.debug("n_unreachable = %d" % n_unreachable)
    size, resident = print_meminfo("AFTER DEL PAIRS")
    size_list.append(size)
    resident_list.append(resident)

  return max(size_list), max(resident_list), failed_ii


def main(argv):
  LOG_LEVEL = logging.INFO
  logging.basicConfig(level=LOG_LEVEL)
  SEQ_LIST_LEN_RANGE = xrange(1000, 10001, 1000)

  try:
    refseq_fn = argv[1]
    read_fn = argv[2]
    mate_fn = argv[3]
  except IndexError:
    sys.exit("Usage: %s REFSEQ_FN READ_FN MATE_FN" % sys.argv[0] + __doc__)
  #u.build_index(refseq_fn)
  for seq_list_len in SEQ_LIST_LEN_RANGE:
    print_meminfo("\nSTART seq_list_len=%d" % seq_list_len, logging.info)
    max_size, max_resident, failed_ii = run_bwa_py_sampe(
      refseq_fn, read_fn, mate_fn, seq_list_len=seq_list_len
      )
    print_meminfo("END seq_list_len=%d" % seq_list_len, logging.info)
    logging.info(
      "PEAK seq_list_len=%d: size=%.1f, resident=%.1f, failed_ii=%d" %
                 (seq_list_len, max_size, max_resident, failed_ii))


if __name__ == "__main__":
  main(sys.argv)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import os

def get_ref_archive(jobconf):
  cwd = os.getcwd()
  archive = jobconf.get("mapred.cache.archives")
  if len(archive.split(",")) > 1:
    raise ValueError("mapred.cache.archives can now only reference one index archive")

  try:
    symlink = archive.split("#")[1].strip()
  except IndexError:
    raise ValueError("Cache entry in mapred.cache.archives MUST be <hdfs_path>#<symlink>")

  return os.path.join(cwd, symlink)




import re
import array

from bl.lib.seq.aligner.mapping import Mapping

class SAMMapping(Mapping):
	"""
	A mapping implementation for storing SAM data.

	A SAMMapping object is constructed from a list of SAM fields --
	see http://samtools.sourceforge.net
	"""

	CIGAR_PATTERN = re.compile(r"(\d+)([MIDNSHP])")
	REFID_PATTERN = re.compile(r"\d+$")
	
	def __init__(self, sam_fields):
		"""
		Provide a sam record as a string (it will be split on tabs to get the
		fields) or directly a list of sam fields.
		"""
		super(SAMMapping, self).__init__()
		if type(sam_fields) == str:
			sam_fields = sam_fields.split("\t")
		self.__name = sam_fields[0]
		self.flag = int(sam_fields[1])
		ref_id_match = self.REFID_PATTERN.search(sam_fields[2]) # it's the best we can do without a sam header or the reference annotations
		if ref_id_match is not None:
			self.ref_id = int(ref_id_match.group())
		self.tid = sam_fields[2]
		self.pos = int(sam_fields[3])
		self.qual = int(sam_fields[4])
		self.__cigar = [(int(n), c) for (n, c) in self.CIGAR_PATTERN.findall(sam_fields[5])]
		if sam_fields[6] == '*':  # is this BWA-specific?
			self.mtid = None
		else:
			self.mtid = sam_fields[6]
		self.mpos = int(sam_fields[7])
		self.isize = int(sam_fields[8])
		self.__seq = sam_fields[9]
		self.__ascii_base_qual = sam_fields[10]
		self.__tags = [tuple(t.split(":", 2)) for t in sam_fields[11:]]

	def get_name(self):
		return self.__name

	def get_seq_5(self):
		return self.__seq

	def get_base_qualities(self):
		if not hasattr(self, '__base_qual'):
			self.__base_qual = array.array(
			  'B', [ord(q) - 33 for q in self.__ascii_base_qual]
			  )
		return self.__base_qual

	def get_cigar(self):
		return self.__cigar

	def each_tag(self):
		for t in self.__tags:
			yield t



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import re
import unittest
from bl.mr.test_utils import map_context, SavingLogger

import bl.lib.seq.aligner.io.protobuf_mapping as proto
from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink
from bl.mr.lib.hadoop_event_monitor import HadoopEventMonitor
from bl.mr.seq.seqal.mapper import MarkDuplicatesEmitter
from bl.mr.seq.seqal.seqal_app import PAIR_STRING, UNMAPPED_STRING
from bl.lib.seq.aligner.sam_mapping import SAMMapping
import test_utils # specific to seqal

class TestMarkDuplicatesEmitter(unittest.TestCase):

	# mini object to let us peek at what the filter forwards to the next link
	class Receiver(HitProcessorChainLink):
		def __init__(self, *args):
			super(type(self), self).__init__(*args)
			self.received = None

		def process(self, pair):
			self.received = pair

	def setUp(self):
		self.ctx = map_context(None, None)
		self.count_group = "Test"
		self.logger = SavingLogger()
		self.monitor = HadoopEventMonitor(self.count_group, self.logger, self.ctx)
		self.link = MarkDuplicatesEmitter(self.ctx, self.monitor)
		self.pair1 = test_utils.pair1()
		self.pair2 = test_utils.pair2()

	def test_forward(self):
		# see whether it forwads the pair to the next link in the chain
		receiver = type(self).Receiver()
		self.link.set_next(receiver)
		self.link.process(self.pair1)
		self.assertFalse(receiver.received is None)
		self.assertEqual(self.pair1, receiver.received)

	def test_forward_reversed(self):
		# see whether a "reversed" pair is emitted unmodified to the next link
		receiver = type(self).Receiver()
		self.link.set_next(receiver)
		self.link.process(self.pair2)
		self.assertFalse(receiver.received is None)
		self.assertEqual(self.pair2, receiver.received)

	def test_emit_forward_pair(self):
		# We expect to get the pair emitted with the key generated from
		# read 1 (the left read).  On the other hand, we expect to get
		# PAIR_STRING with the key generated from read 2
		self.link.process(self.pair1)
		expected_keys = map(test_utils.make_key, self.pair1)
		for i in 0,1:
			self.assertTrue( self.ctx.emitted.has_key(expected_keys[i]) )
			self.assertEqual(1, len(self.ctx.emitted[expected_keys[i]]))

		unserialized = proto.unserialize_pair(self.ctx.emitted[expected_keys[0]][0])
		for j in 0,1:
			self.assertEqual(self.pair1[j].tid, unserialized[j].tid)
			self.assertEqual(self.pair1[j].pos, unserialized[j].pos)

		second_value = self.ctx.emitted[expected_keys[1]][0]
		self.assertEqual(PAIR_STRING, second_value)

	def test_emit_backward_pair(self):
		# Similar to test_emit_forward_pair, but here we expect to have the reads
		# reordered.  So,
		#   key2 => reversed and serialized pair
		#   key1 => PAIR_STRING
		self.link.process(self.pair2)
		expected_keys = map(test_utils.make_key, self.pair2)
		for i in 0,1:
			self.assertTrue( self.ctx.emitted.has_key(expected_keys[i]) )
			self.assertEqual(1, len(self.ctx.emitted[expected_keys[i]]))

		unserialized = proto.unserialize_pair(self.ctx.emitted[expected_keys[1]][0])
		for j in 0,1:
			self.assertEqual(self.pair2[j].tid, unserialized[j^1].tid)
			self.assertEqual(self.pair2[j].pos, unserialized[j^1].pos)

		second_value = self.ctx.emitted[expected_keys[0]][0]
		self.assertEqual(PAIR_STRING, second_value)

	def test_count_emitted_records(self):
		# we expect to get the pair emitted twice, once with the key generated from
		# read 1 and once with the read generated from read 2
		self.link.process(self.pair1)
		self.assertTrue(self.ctx.counters.has_key("Test:MAPPED COORDINATES"))
		self.assertEqual(2, self.ctx.counters["Test:MAPPED COORDINATES"])

	def test_emit_forward_fragment1(self):
		# None in pair[0]. Fragment in pair[1].
		self.pair1 = test_utils.erase_read1(list(self.pair1))
		self.link.process(self.pair1)
		self.assertEqual(1, len(self.ctx.emitted.keys()))
		expected_key = test_utils.make_key(self.pair1[1])
		self.assertEqual(1, len(self.ctx.emitted[expected_key]))
		unserialized = proto.unserialize_pair(self.ctx.emitted[expected_key][0])
		self.assertTrue(unserialized[1] is None)
		self.assertEqual(self.pair1[1].tid, unserialized[0].tid)
		self.assertEqual(self.pair1[1].pos, unserialized[0].pos)
		self.assertTrue(self.ctx.counters.has_key("Test:MAPPED COORDINATES"))
		self.assertEqual(1, self.ctx.counters["Test:MAPPED COORDINATES"])

	def test_emit_forward_fragment2(self):
		# Fragment in pair[0].  None in pair[1]
		self.pair1 = test_utils.erase_read2(list(self.pair1))
		self.link.process(self.pair1)
		self.assertEqual(1, len(self.ctx.emitted.keys()))
		expected_key = test_utils.make_key(self.pair1[0])
		self.assertEqual(1, len(self.ctx.emitted[expected_key]))
		unserialized = proto.unserialize_pair(self.ctx.emitted[expected_key][0])
		self.assertTrue(unserialized[1] is None)
		self.assertEqual(self.pair1[0].tid, unserialized[0].tid)
		self.assertEqual(self.pair1[0].pos, unserialized[0].pos)
		self.assertTrue(self.ctx.counters.has_key("Test:MAPPED COORDINATES"))
		self.assertEqual(1, self.ctx.counters["Test:MAPPED COORDINATES"])

	def test_emit_reverse_fragment1(self):
		# None in pair[0]. Fragment in pair[1].
		self.pair1 = test_utils.erase_read1(list(self.pair1))
		self.pair1[1].set_on_reverse(True)
		self.link.process(self.pair1)
		self.assertEqual(1, len(self.ctx.emitted.keys()))
		expected_key = test_utils.make_key(self.pair1[1])
		self.assertEqual(1, len(self.ctx.emitted[expected_key]))
		unserialized = proto.unserialize_pair(self.ctx.emitted[expected_key][0])
		self.assertTrue(unserialized[1] is None)
		self.assertEqual(self.pair1[1].tid, unserialized[0].tid)
		self.assertEqual(self.pair1[1].pos, unserialized[0].pos)
		self.assertTrue(unserialized[0].is_on_reverse())

	def test_unmapped1(self):
		self.pair1[0].set_mapped(False)
		self.pair1[1].set_mate_mapped(False)
		self.link.process(self.pair1)

		self.assertEqual(1, len(self.ctx.emitted.keys()))
		self.assertTrue( test_utils.make_key(self.pair1[1]) in self.ctx.emitted.keys() )
		self.assertEqual(1, len(self.ctx.emitted.values()[0]))
		unserialized = proto.unserialize_pair(self.ctx.emitted.values()[0][0])
		self.assertFalse(unserialized[0] is None)
		self.assertFalse(unserialized[1] is None)
		self.assertEqual(self.pair1[1].tid, unserialized[0].tid)
		self.assertEqual(self.pair1[1].pos, unserialized[0].pos)
		self.assertEqual(1, self.ctx.counters["Test:UNMAPPED READS"])

	def test_unmapped2(self):
		self.pair1[1].set_mapped(False)
		self.pair1[0].set_mate_mapped(False)
		self.link.process(self.pair1)
		self.assertEqual(1, len(self.ctx.emitted.keys()))
		self.assertEqual(0, len(filter(lambda k: re.match(UNMAPPED_STRING + ":\d+", k), self.ctx.emitted.keys())))
		self.assertTrue( test_utils.make_key(self.pair1[0]) in self.ctx.emitted.keys() )
		self.assertEqual(1, self.ctx.counters["Test:UNMAPPED READS"])

	def test_both_unmapped(self):
		for i in 0,1:
			self.pair1[i].set_mapped(False)
			self.pair1[i].set_mate_mapped(False)
		self.link.process(self.pair1)
		self.assertEqual(1, len(self.ctx.emitted.keys()))
		self.assertEqual(1, len(filter(lambda k: re.match(UNMAPPED_STRING + ":\d+", k), self.ctx.emitted.keys())))
		self.assertEqual(2, self.ctx.counters["Test:UNMAPPED READS"])

	def test_rmdup_bug(self):
		test_case_data = [
"HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t83\t20\t6181935\t60\t5S96M\t=\t6181919\t-112\tAAGTGGAAGATTTGGGAATCTGAGTGGATTTGGTAACAGTAGAGGGGTGGATCTGGCTTGGAAAACAATCGAGGTACCAATATAGGTGGTAGATGAATTTT\t?<?AADADBFBF<EHIGHGGGEAF3AF<CHGGDG9?GHFFACDHH)?@AHEHHIIIIE>A=A:?);B27@;@?>,;;C(5::>>>@5:()4>@@@######\tXC:i:96\tXT:A:U\tNM:i:1\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:1\tXO:i:0\tXG:i:0\tMD:Z:13G82\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:######@@@>4)(:5@>>>::5(C;;,>?@;@72B;)?:A=A>EIIIIHHEHA@?)HHDCAFFHG?9GDGGHC<FA3FAEGGGHGIHE<FBFBDADAA?<?",
"HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t163\t20\t6181919\t60\t101M\t=\t6181935\t112\tCTGAGCACACCAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACACCTCTACTGTTACCAAATCCACTCAGATTCCCAA\t@@@FFFDDFHG??;EEH>HHGIGHEGCGEGGIGJG31?DDBBD>FGG@HG??DFBBADFAGII3@EH;;CEHECBB7?>CE.;...5>ACDDA:C:;>:>?\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:29G36C34\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:@@@FFFDDFHG??;EEH>HHGIGHEGCGEGGIGJG31?DDBBD>FGG@HG??DFBBADFAGII3@EH;;CEHECBB7?>CE.;...5>ACDDA:C:;>:>?",
"HWI-ST200R_251:6:2207:18561:163438#GCCAAT\t83\t20\t6181938\t60\t8S93M\t=\t6181919\t-112\tAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACCCCTCTACTGTTACCAAATCCACTCAGATTCCCAAATCTTCCACTT\t@@@DDFDFHHHHHJJJEHGGHIHHAEGHJJIJJFGGHGIDIGIJJ?BBGGGIIIJJIJGFHGIJEC(=3?C;?B9?@C>CECECAA(;;3>C#########\tXC:i:93\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:10G36C45\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:#########C>3;;(AACECEC>C@?9B?;C?3=(CEJIGHFGJIJJIIIGGGBB?JJIGIDIGHGGFJJIJJHGEAHHIHGGHEJJJHHHHHFDFDD@@@",
"HWI-ST200R_251:6:2207:18561:163438#GCCAAT\t163\t20\t6181919\t60\t101M\t=\t6181938\t112\tCTGAGCACACCAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACACCTCTACTGTTACCAAATCCACTCAGATTCCCAA\t@CCFFDDFHHHHHIJJJIIJJJIJJIIJGJIIIJII?DGHIGHDGHIIIJIJIJIIDCHGIJIIGGHIFEHHHHFFFFFDC.6.66;@CCCDCCDC>CCCA\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:29G36C34\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:@CCFFDDFHHHHHIJJJIIJJJIJJIIJGJIIIJII?DGHIGHDGHIIIJIJIJIIDCHGIJIIGGHIFEHHHHFFFFFDC.6.66;@CCCDCCDC>CCCA",
		]
		sams = map(SAMMapping, test_case_data)
		pair1 = sams[0:2]
		pair2 = sams[2:]
		self.link.process(pair1)

		self.assertEqual(2, len(self.ctx.emitted.keys()))
		key_list = list(sorted(self.ctx.emitted.keys()))
		self.assertEqual("0020:000006181919:F", key_list[0])
		self.assertEqual("0020:000006182030:R", key_list[1])

		self.link.process(pair2)

		self.assertEqual(2, len(self.ctx.emitted.keys()))
		for k,value_list in self.ctx.emitted.iteritems():
			self.assertEqual(2, len(value_list)) # each key should have two pairs at its position

		value_list = self.ctx.emitted["0020:000006181919:F"]
		for value in value_list:
			unserialized = proto.unserialize_pair(value)
			self.assertTrue(unserialized[0].pos < unserialized[1].pos)
		value_list = self.ctx.emitted["0020:000006182030:R"]
		for value in value_list:
			self.assertEqual(PAIR_STRING, value)

	def test_fw_rev_with_indels(self):
		"""
		Here we have two duplicate pairs.  Read 1 in both pairs is positioned at the same location.
		For both pairs, read 2 is mapped on the reverse strand.  The second one is mapped with an
		insertion, so its 5' location is shifted by one.  Yet, the read end is in the same location.
		"""
		test_case_data = [
"HWI-ST200R_251:7:2207:3236:93050#CGATGT\t99\t7\t15609040\t60\t101M\t=\t15609197\t257\tCTAGCTTGTAACAATTGCTATAACTCCCCCACTTTGGATGGTAAATTTCTCCTCAGCTGTCATTGGCCCTCAAAGCCAAAATGACTCCAATTAGAATGTAT\tCCCFFFFFHHHHHJJJJJJJJJIIJJJJJJJIJJJJJJJJJFIJJJJJJJJJJJIJJIJGIGIJJJJJJJJJ>EHHEFFFFFEEDEDEDDCDDDACCA:CD",
"HWI-ST200R_251:7:2207:3236:93050#CGATGT\t147\t7\t15609197\t60\t37M1I63M\t=\t15609040\t-257\tCTCCATTACAGCAGAGGAAAGAAACTTTTTTTTTTTCTTTTTTTTTTTTTTTTTTTAAAGAAACTGGGTTGAAGAAGTAGTTCATTGAATGGTTGTCTTAC\t################CAC@3DA:))&BDDDDDDB<&BDDDDDDDDHIJJIJJIJJJJJIIHHJJJJJJJJJJJJJJIJJJJJJJJJJHHHHHFFFFFBBC",
"HWI-ST200R_251:1:1101:10006:13364#CGATGT\t99\t7\t15609040\t60\t101M\t=\t15609196\t257\tCTAGCTTGTAACAATTGCTATAACTCCCCCACTTTGGATGGTAAATTTCTCCTCAGCTGTCATTGGCCCTCAAAGCCAAAATGACTCCAATTAGAATGTAT\tCCCFFFFFHHHHHJJJJJJJJJJJGIJJJJIJJJJJJJJJIHIIJJJJJJJJJJJJIIJGGGIIIJDHIGIHFGFEHEF>??>@CDEECC@CCC(>;A:>5",
"HWI-ST200R_251:1:1101:10006:13364#CGATGT\t147\t7\t15609196\t60\t101M\t=\t15609040\t-257\tTACCATTTAAAGCAGAGGAAAAAAACTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTAAAGAAACTGGGTTGAAGAAGTAGTTCATTGAATGGTTGTCTTAC\t############################BBDDDDB803DDDDDDDDHJJJIIJJJJJJJIHF?JJJJJJJJJIIJJJJJJJJJJJJJJHHHHHFFFFFB@C",
		]

		sams = map(SAMMapping, test_case_data)
		pair1 = sams[0:2]
		pair2 = sams[2:]
		self.link.process(pair1)
		self.link.process(pair2)

		self.assertEqual(2, len(self.ctx.emitted.keys()))
		key_list = list(sorted(self.ctx.emitted.keys()))
		self.assertEqual("0007:000015609040:F", key_list[0])
		self.assertEqual("0007:000015609296:R", key_list[1])

		for k,value_list in self.ctx.emitted.iteritems():
			self.assertEqual(2, len(value_list)) # each key should have two pairs at its position

		value_list = self.ctx.emitted["0007:000015609040:F"]
		for value in value_list:
			unserialized = proto.unserialize_pair(value)
			self.assertTrue(unserialized[0].pos < unserialized[1].pos)
		value_list = self.ctx.emitted["0007:000015609296:R"]
		for value in value_list:
			self.assertEqual(PAIR_STRING, value)

	def test_fw_rev_missed_dup_pair(self):
		"""
		Here we have two duplicate pairs, as detected by Picard.
		The reads 10364 have flags 'pPR1', 'pPr2'.
		The reads 138222 have flags 'pPR2d', 'pPr1d'.
		We expect a key position of 88888339:F for both pairs.
		"""
		test_case_data = [
			"HWI-ST332_97:3:7:10556:138222#0\t83\t10\t88888404\t23\t5S71M\t=\t88888399\t-76\tTGATGTTGCTCCATTGTCTTCTAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCATTCTGGAAGCTTATCTATTG\t######C<BCC:B,DDDC=BD3@CB8B?DBD@E@EEEEECED@CDB=8C7A@D=DEDEDCDBECDE<>;;,17,45",
			"HWI-ST332_97:3:7:10556:138222#0\t163\t10\t88888399\t15\t50M26S\t=\t88888404\t76\tTGATTTTGCTCCATTGTCTTCTAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCCGTCTGGTTGCTTCTATTTTG\tCC7EEFFF@FHHFHHG?GGF:>4.7GD8DC@D>CCFGG@GGG5GG4<CB###########################",
			"HWI-ST332_97:3:66:16214:10364#0\t99\t10\t88888399\t46\t76M\t=\t88888421\t76\tTGATTTTGCTCCATTGTCTTCTAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCATTCTGGAAGCTTATCTATTG\tHGHHHHHHFHEHHHHGHHFHHHGGHHHHHHHHHHEHDHGEHFHHHHHHHGGHHHHHHHHHHBHBBEGG=GFFFF@F",
			"HWI-ST332_97:3:66:16214:10364#0\t147\t10\t88888421\t46\t22S54M\t=\t88888399\t-76\tTGATTCCGCTCCATGTGCCTCGAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCATTCTGGAAGCTTATCTATTG\t#######################AA@EHGHEHHHHHHHHHHHHHHHEHHHHHHFHFHHHHHHHHHHHHHHHGHHFH",
		]
		sams = map(SAMMapping, test_case_data)
		pair1 = sams[0:2]
		pair2 = sams[2:]
		self.link.process(pair1)

		self.assertEqual(2, len(self.ctx.emitted.keys()))
		key_list = list(sorted(self.ctx.emitted.keys()))
		self.assertEqual("0010:000088888399:F", key_list[0])
		self.assertEqual("0010:000088888474:R", key_list[1])

		self.link.process(pair2)
		self.assertEqual(2, len(self.ctx.emitted.keys()))
		key_list = list(sorted(self.ctx.emitted.keys()))
		self.assertEqual("0010:000088888399:F", key_list[0])
		self.assertEqual("0010:000088888474:R", key_list[1])

		# ensure all pairs are emitted with the first key
		for value in self.ctx.emitted["0010:000088888474:R"]:
			self.assertEqual(PAIR_STRING, value)


def suite():
	return unittest.TestLoader().loadTestsFromTestCase(TestMarkDuplicatesEmitter)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

from bl.lib.seq.aligner.sam_flags import *

from itertools import izip
import array

class Mapping(object):
	"""
	Abstract class defining an interface to describe
	sequence mapping.  A concrete implementation needs
	to set the attributes to appropriate values and
	implement any abstract methods (e.g. get_seq_5).

	Attributes:
	self.flag: SAM flag bits
	self.isize: insert size
	self.mpos: mate's position, or 0
	self.mtid: mate's reference , or '=' if same as query's reference, or None
	self.m_ref_id:  mate's reference id, or None
	self.pos: this hit's position, or 0
	self.qual: mapq value for this alignment
	self.tid: this hit's reference sequence, or None
	self.ref_id:  this hit's reference id, or None
	"""

	def __init__(self):
		self.flag = 0
		self.isize = 0
		self.mpos = 0
		self.mtid = None
		self.m_ref_id = None
		self.pos = 0
		self.qual = 0
		self.tid = None
		self.ref_id = None

	def get_name(self):
		"""This fragment's id"""
		abstract

	def get_seq_5(self):
		abstract

	def get_seq_len(self):
		return len(self.get_seq_5())

	def get_base_qualities(self):
		"""If the base qualities are available, this method
		returns a byte array.array of Phred quality scores,
		one per base, 0-based.  If base qualities aren't
		available it return None"""
		abstract

	def get_ascii_base_qual(self):
		"""
		Returns a string containing the Phred base quality score
		encoded in ASCII-33 (ASCII character corresponding to
		quality score + 33)
		"""
		if not hasattr(self, '__ascii_base_qual'):
			self.__ascii_base_qual = ''.join([ chr(q+33) for q in self.get_base_qualities() ])
		return self.__ascii_base_qual

	def get_cigar(self):
		"""Return the cigar list for this mapping.
		Each element is a pair (length, action),
		where action is a letter for one of the alignment actions
		(e.g. M, I, D, S).  If unmapped return an empty list.
		"""
		abstract

	def get_cigar_str(self):
		if not hasattr(self, '__cigar_str'):
			self.__cigar_str = "".join(['%d%s' % t for t in self.get_cigar()]) or "*"
		return self.__cigar_str

	def get_untrimmed_left_pos(self):
		"""
		For a reversed, trimmed fragment the 5' coordinate pos refers to a base
		somewhere within the fragment (not at its start or end).  This method
		takes into account the trimming (if any, shown as soft or hard clipping
		in the CIGAR) and calculates the pos of the 5'-most base of the untrimmed
		fragment.
		"""
		if not hasattr(self, "__untrimmed_left_pos"):
			if self.is_unmapped():
				raise ValueError("sequence is not mapped")

			upos = self.pos
			for length, op in self.get_cigar():
				if op == 'S' or op == 'H':
					upos -= length
				else:
					break
			self.__untrimmed_left_pos = upos
		return self.__untrimmed_left_pos

	def get_right_pos(self):
		"""
		This computes the reference position of the right-most base of the sequence.
		"""
		if not hasattr(self, "__right_pos"):
			if self.is_unmapped():
				raise ValueError("sequence is not mapped")

			advancement = 0
			advancing_ops = ('M','D','N')
			for length, op in self.get_cigar():
				if op in advancing_ops:
					advancement += length
			if advancement > 0:
				advancement -= 1
			self.__right_pos = self.pos + advancement
		return self.__right_pos

	def get_untrimmed_right_pos(self):
		"""
		This computes the reference position of the right-most base of the sequence,
		including any trimmed bases.
		"""
		right_pos = self.get_right_pos()
		for length, op in reversed(self.get_cigar()):
			if op == 'S' or op == 'H':
				right_pos += length
			else:
				break
		return right_pos


	####################################
	# Tag methods
	####################################
	def each_tag(self):
		"""Generator for this mapping's tags.  Each tag is
		a tuple (tag_name, value_type, value)"""
		abstract

	def tag_value(self, tag_name):
		"""
		Get the value for a tag, if present.
		Returns:
			- integer (if tag type is 'i')
			- float (if tag type is 'f')
			- string
			- None, if tag isn't present
		"""
		for tag in self.each_tag():
			if tag[0] == tag_name:
				if tag[1] == 'i':
					value = int(tag[2])
				elif tag[1] == 'f':
					value = float(tag[2])
				else:
					value = tag[2]
				return value
		return None

	####################################
	# Flag methods
	####################################
	def flag_string(self):
		"""
			Returns a string indicating symbolically the flags set for this mapping.
			Symbol legend:
			p: paired										FPD	 0x001
			P: properly paired					 FPP	 0x002
			u: query unmapped						FSU	 0x004
			U: mate unmapped						 FMU	 0x008
			r: query on reverse strand	 FSR	 0x010
			R: mate on reverse strand		FMR	 0x020
			1: first read in a pair			FR1	 0x040
			2: second read in a pair		 FR2	 0x080
			s: not a primary alignment	 FSC	 0x100
			f: fails quality checks						0x200
			d: duplicate											 0x400
		"""
		names = [ 'p', 'P', 'u', 'U', 'r', 'R', '1', '2', 's', 'f', 'd']
		values = [ 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080, 0x0100, 0x0200, 0x0400]
		return ''.join( [ n for n,v in izip(names, values) if self.flag & v != 0 ] )


	def is_paired(self):
		return self.flag & SAM_FPD != 0

	def is_properly_paired(self):
		return self.flag & SAM_FPP != 0

	def is_mapped(self):
		return self.flag & SAM_FSU == 0

	def is_unmapped(self):
		return self.flag & SAM_FSU != 0

	def is_mate_mapped(self):
		return self.flag & SAM_FMU == 0

	def is_mate_unmapped(self):
		return self.flag & SAM_FMU != 0

	def is_on_reverse(self):
		return self.flag & SAM_FSR != 0

	def is_mate_on_reverse(self):
		return self.flag & SAM_FMR != 0

	def is_read1(self):
		return self.flag & SAM_FR1 != 0

	def is_read2(self):
		return self.flag & SAM_FR2 != 0

	def is_secondary_align(self):
		return self.flag & SAM_FSC != 0

	def is_failed_qc(self):
		return self.flag & SAM_FQC != 0

	def is_duplicate(self):
		return self.flag & SAM_FDP != 0

	def set_paired(self, value):
		self.__set_flag(value, SAM_FPD)

	def set_properly_paired(self, value):
		self.__set_flag(value, SAM_FPP)

	def set_mapped(self, value):
		self.__set_flag(not value, SAM_FSU)

	def set_mate_mapped(self, value):
		self.__set_flag(not value, SAM_FMU)

	def set_on_reverse(self, value):
		self.__set_flag(value, SAM_FSR)

	def set_mate_on_reverse(self, value):
		self.__set_flag(value, SAM_FMR)

	def set_read1(self, value):
		self.__set_flag(value, SAM_FR1)

	def set_read2(self, value):
		self.__set_flag(value, SAM_FR2)

	def set_secondary_align(self, value):
		self.__set_flag(value, SAM_FSC)

	def set_failed_qc(self, value):
		self.__set_flag(value, SAM_FQC)

	def set_duplicate(self, value):
		self.__set_flag(value, SAM_FDP)

	def __set_flag(self, new_value, flag_mask):
		if new_value:
			self.flag |= flag_mask
		else:
			self.flag &= ~flag_mask

	def remove_mate(self):
		"""
		removes all info pertaining to this mapping's mate.
		"""
		self.set_properly_paired(False)
		self.set_mate_mapped(True) # turns off the bit
		self.mpos = 0
		self.mtid = None
		self.m_ref_id = None
		self.isize = 0

######################################
# SimpleMapping
# A mapping implementation with setters for  attributes.
# Useful for testing.
######################################
class SimpleMapping(Mapping):
	def __init__(self):
		super(SimpleMapping, self).__init__()
		self.__name = ""
		self.__base_qualities = array.array('B')
		self.__cigar = []
		self.__tags = []
		self.__seq = ""

	def set_name(self, name):
		self.__name = name

	def set_base_qualities(self, bq):
		if bq is not None and not isinstance(bq, array.array):
			raise TypeError("Invalid type for base qualities (%s)" % str(type(bq)))
		self.__base_qualities = bq

	def set_cigar(self, cig):
		if cig is not None and not isinstance(cig, list):
			raise TypeError("Invalid type for cigar (%s)" % str(type(cig)))
		self.__cigar = cig

	def add_tag(self, tag):
		self.__tags.append(tag)

	def set_seq_5(self, s):
		self.__seq = s

	def get_name(self):
		return self.__name

	def get_base_qualities(self):
		return self.__base_qualities

	def get_cigar(self):
		return self.__cigar

	def each_tag(self):
		for t in self.__tags:
			yield t

	def get_seq_5(self):
		return self.__seq

	def get_seq_len(self):
		return len(self.__seq)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


import bl.lib.seq.aligner.io.mapping_pb2 as proto
from bl.lib.seq.aligner.mapping import Mapping

import re
import array

def serialize_pair(mapping_pair):
	if len(mapping_pair) != 2:
		raise ValueError("Pair must have 2 mappings. (it had %d)" % len(mapping_pair))
	if not any(mapping_pair):
		raise ValueError("No mappings found in pair")

	message = proto.PairMapping()
	isizes = [0,0]
	names = [None, None]
	for idx, ltr in zip( (0,1), ('A','B') ):
		if mapping_pair[idx]:
			# write mapping_pair[idx] to message.readA or message.readB
			__seq_to_protobuf(getattr(message, "read" + ltr), mapping_pair[idx])
			isizes[idx] = mapping_pair[idx].isize
			names[idx] = re.sub("/[12]$", "", mapping_pair[idx].get_name())

	if all(mapping_pair): # a real pair
		if isizes[0] != -1*isizes[1]:
			raise ValueError("pair insert sizes don't match (%d != -1*%d)" % tuple(isizes))
		message.insert_size = abs(isizes[0])
		if names[0] != names[1]:
			raise ValueError("pair query ids don't match (%s != %s)" % tuple(names))
		message.query_id = names[0]
		message.readA.ClearField("query_id")
		message.readB.ClearField("query_id")
	else: # only one mapping
		message.query_id = names[0] if mapping_pair[0] is not None else names[1] # pick the one that isn't None

	return message.SerializeToString()

def unserialize_pair(message):
	incoming = proto.PairMapping()
	incoming.ParseFromString(message)

	pair = [None, None]
	if incoming.HasField("readA"):
		# We cleared the query_ids when we serialized the pair. Here we restore them
		incoming.readA.query_id = incoming.query_id
		pair[0] = UnserializedMapping()
		pair[0].set_from_memento(incoming.readA)
	if incoming.HasField("readB"):
		incoming.readB.query_id = incoming.query_id
		pair[1] = UnserializedMapping()
		pair[1].set_from_memento(incoming.readB)

	if all(pair): # a real pair
		# The insert size isn't stored in the SeqMapping, so we reset it
		# here.  The mapping that is further up on the chromosome gets
		# the negative insert size
		pair[0].isize = pair[1].isize = incoming.insert_size
		if pair[0].pos < pair[1].pos:
			pair[1].isize *= -1
		else:
			pair[0].isize *= -1
		# set mate position
		for i in 0,1:
			one = pair[i]
			other = pair[i^1]
			one.m_ref_id = other.ref_id
			one.mtid = other.tid
			one.mpos = other.pos

	return tuple(pair)

def serialize_seq(mapping):
	message = proto.SeqMapping()
	return __seq_to_protobuf(message, mapping).SerializeToString()

def unserialize_seq(message):
	incoming = proto.SeqMapping()
	incoming.ParseFromString(message)
	mapping = UnserializedMapping()
	mapping.set_from_memento(incoming)
	return mapping

##
# Theis private function does the actual conversion
# from Mapping object to protobuf SeqMapping message.
def __seq_to_protobuf(proto_map, mapping):
	proto_map.query_id = mapping.get_name()
	proto_map.flags = mapping.flag
	proto_map.reference = mapping.tid or "*"
	proto_map.reference_id = mapping.ref_id if mapping.ref_id is not None else -1
	proto_map.position = mapping.pos
	proto_map.map_q = mapping.qual
	proto_map.cigar = mapping.get_cigar_str()

	proto_map.sequence = mapping.get_seq_5()
	proto_map.sanger_qual = mapping.get_base_qualities().tostring()

	for tag_name, tp, value in mapping.each_tag():
		new_tag = proto_map.tags.add()
		new_tag.name = tag_name
		if tp == "i":
			new_tag.type = proto.SeqMapping.Tag.INTEGER
			new_tag.int_value = int(value)
		elif tp == "f":
			new_tag.type = proto.SeqMapping.Tag.FLOAT
			new_tag.double_value = float(value)
		elif tp == "A":
			new_tag.type = proto.SeqMapping.Tag.CHARACTER
			new_tag.other_value = value
		elif tp == "H":
			new_tag.type = proto.SeqMapping.Tag.HEXSTRING
			new_tag.other_value = value
		elif tp == "Z":
			new_tag.type = proto.SeqMapping.Tag.STRING
			new_tag.other_value = value
		else:
			raise ValueError("Unknown tag type " + tp)
	return proto_map

class UnserializedMapping(Mapping):
	def __init__(self):
		super(UnserializedMapping, self).__init__()

	def get_name(self):
		return self.__name

	def get_base_qualities(self):
		return self.__base_qualities

	def get_cigar(self):
		return self.__cigar

	def each_tag(self):
		for t in self.__tags:
			yield t

	def get_seq_5(self):
		return self.__seq

	def get_seq_len(self):
		return len(self.__seq)

	def set_from_memento(self, protobuf_seq_mapping):
		if protobuf_seq_mapping.HasField("sanger_qual"):
			self.__base_qualities = array.array("B")
			self.__base_qualities.fromstring(protobuf_seq_mapping.sanger_qual)
		self.__cigar = UnserializedMapping.__scan_cigar(protobuf_seq_mapping.cigar)
		self.flag = protobuf_seq_mapping.flags
		self.__name = protobuf_seq_mapping.query_id
		self.qual = protobuf_seq_mapping.map_q
		self.pos = protobuf_seq_mapping.position
		self.__tags = UnserializedMapping.__build_tags(protobuf_seq_mapping.tags)
		self.tid = protobuf_seq_mapping.reference if protobuf_seq_mapping.reference != "*" else None
		self.ref_id = protobuf_seq_mapping.reference_id if protobuf_seq_mapping.reference_id > -1 else None
		self.__seq = protobuf_seq_mapping.sequence

		# sanity checks
		if self.__base_qualities and self.__seq != "=":
			if len(self.__base_qualities) != len(self.__seq):
				raise RuntimeError("sequence length (%d) differs from base_qualities length (%d)" %
					(len(self.__seq), len(self.__base_qualities)))

	@staticmethod
	def __scan_cigar(cig_str):
		if cig_str == "*":
			return []
		else:
			# verify syntax
			if not re.match("(\d+[MIDS])+", cig_str):
				raise ValueError("Invalid CIGAR syntax: " + cig_str)
			return [ (int(m.group(1)), m.group(2)) for m in re.finditer("(\d+)([MIDS])", cig_str) ]

	@staticmethod
	def __build_tags(serialized):
		tags = []
		for ptag in serialized:
			if ptag.type == proto.SeqMapping.Tag.INTEGER:
				ttype = 'i'
				tvalue = ptag.int_value
			elif ptag.type == proto.SeqMapping.Tag.FLOAT:
				ttype = 'f'
				tvalue = ptag.double_value
			elif ptag.type ==  proto.SeqMapping.Tag.CHARACTER:
				ttype = 'A'
				tvalue = ptag.other_value
			elif ptag.type == proto.SeqMapping.Tag.HEXSTRING:
				ttype = 'H'
				tvalue = ptag.other_value
			elif ptag.type == proto.SeqMapping.Tag.STRING:
				ttype = "Z"
				tvalue = ptag.other_value
			else:
				raise ValueError("Invalid TagType type: " + ptag.type)
			tags.append( (ptag.name, ttype, tvalue) )

		return tags




# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import re
import time
import unittest

from bl.mr.lib.hadoop_event_monitor import *
from bl.mr.test_utils import map_context, SavingLogger

class TestHadoopEventMonitor(unittest.TestCase):

	def setUp(self):
		self.map_ctx = map_context(None, None)
		self.logger = SavingLogger()
		self.count_group = "Test"
		self.monitor = HadoopEventMonitor(self.count_group, self.logger, self.map_ctx)

	def test_start_stop(self):
		self.monitor.start("test_start_stop")
		time.sleep(0.5)
		self.monitor.stop("test_start_stop")
		self.assertTrue( re.match('done with test_start_stop.*', self.logger.contents()) is not None)
		self.assertTrue( re.match('done with test_start_stop.*', self.map_ctx.status_messages[-1]) is not None)
		# extract the measured time and ensure it's > 0
		m = re.search("done with test_start_stop\s+\(\s*(-?[0-9.]+\s*)", self.map_ctx.status_messages[-1])
		self.assertTrue(m is not None, "couldn't extract time from timing status message")
		self.assertTrue(float(m.group(1)) > 0)

	def test_time_block(self):
		with self.monitor.time_block("test_time_block"):
			# in block
			time.sleep(0.5)

		self.assertTrue( re.match('done with test_time_block.*', self.logger.contents()) is not None)
		self.assertTrue( re.match('done with test_time_block.*', self.map_ctx.status_messages[-1]) is not None)
		# extract the measured time and ensure it's > 0
		m = re.search("done with test_time_block\s+\(\s*(-?[0-9.]+\s*)", self.map_ctx.status_messages[-1])
		self.assertTrue(m is not None, "couldn't extract time from timing status message")
		self.assertTrue(float(m.group(1)) > 0)

	def test_count_default(self):
		# this corresponds to how the mock context creates the counter ids,
		# and the event monitor makes the name uppercase.
		counter_id = "%s:%s" % (self.count_group, "TEST_COUNT")
		self.monitor.count("test_count")
		self.assertEqual(1, self.map_ctx.counters[counter_id])
		self.monitor.count("test_count")
		self.assertEqual(2, self.map_ctx.counters[counter_id])

	def test_count_nondefault_value(self):
		counter_id = "%s:%s" % (self.count_group, "TEST_COUNT")
		self.monitor.count("test_count", 8)
		self.assertEqual(8, self.map_ctx.counters[counter_id])
		self.monitor.count("test_count", 5)
		self.assertEqual(13, self.map_ctx.counters[counter_id])

	def test_count_default_preexisting(self):
		counter_id = "%s:%s" % (self.count_group, "TEST_COUNT")
		self.monitor.add_counter("test_count")
		self.monitor.count("test_count")
		self.assertEqual(1, self.map_ctx.counters[counter_id])
		self.monitor.count("test_count")
		self.assertEqual(2, self.map_ctx.counters[counter_id])

	def test_has_counter_add_counter(self):
		self.assertFalse( self.monitor.has_counter("test_has_counter") )
		self.monitor.count("test_has_counter")
		self.assertTrue( self.monitor.has_counter("test_has_counter") )
		self.assertFalse( self.monitor.has_counter("other_counter") )
		self.monitor.add_counter("other_counter")
		self.assertTrue( self.monitor.has_counter("other_counter") )

	def test_new_status(self):
		self.assertEqual(0, len(self.map_ctx.status_messages))
		self.monitor.new_status("test_new_status")
		self.assertEqual(1, len(self.map_ctx.status_messages))
		self.assertEqual("test_new_status", self.map_ctx.status_messages[-1])
		self.assertEqual("test_new_status\n", self.logger.contents())

	def test_new_status_lower_log_level(self):
		self.logger.log_level = SavingLogger.ERROR
		self.assertEqual(0, len(self.map_ctx.status_messages))
		self.monitor.new_status("test_new_status")
		self.assertEqual(1, len(self.map_ctx.status_messages))
		self.assertEqual("test_new_status", self.map_ctx.status_messages[-1])
		self.assertEqual("", self.logger.contents())

	def test_log_debug(self):
		self.logger.log_level = SavingLogger.DEBUG
		self.monitor.log_debug("debug")
		self.assertEqual("debug\n", self.logger.contents())
		self.monitor.log_info("info")
		self.assertEqual("debug\ninfo\n", self.logger.contents())

	def test_log_info(self):
		self.logger.log_level = SavingLogger.INFO
		self.monitor.log_debug("debug")
		self.assertEqual("", self.logger.contents())
		self.monitor.log_info("info")
		self.assertEqual("info\n", self.logger.contents())
		self.monitor.log_error("error")
		self.assertEqual("info\nerror\n", self.logger.contents())

	def test_log_warning(self):
		self.logger.log_level = SavingLogger.WARNING
		self.monitor.log_info("info")
		self.assertEqual("", self.logger.contents())
		self.monitor.log_warning("warning")
		self.assertEqual("warning\n", self.logger.contents())
		self.monitor.log_error("error")
		self.assertEqual("warning\nerror\n", self.logger.contents())

	def test_log_error(self):
		self.logger.log_level = SavingLogger.ERROR
		self.monitor.log_info("info")
		self.assertEqual("", self.logger.contents())
		self.monitor.log_error("error")
		self.assertEqual("error\n", self.logger.contents())
		self.monitor.log_critical("critical")
		self.assertEqual("error\ncritical\n", self.logger.contents())

	def test_log_critical(self):
		self.logger.log_level = SavingLogger.CRITICAL
		self.monitor.log_info("info")
		self.assertEqual("", self.logger.contents())
		self.monitor.log_critical("critical")
		self.assertEqual("critical\n", self.logger.contents())

def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestHadoopEventMonitor)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())
